from __future__ import annotations

import dataclasses
import itertools
import json
from collections import defaultdict
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Generic, TypeVar, cast

import trio
from trio_websocket import ConnectionClosed, connect_websocket_url

from streamlink.logger import getLogger
from streamlink.webbrowser.cdp.devtools.target import SessionID, attach_to_target, create_target
from streamlink.webbrowser.cdp.devtools.util import CDPEvent, parse_json_event
from streamlink.webbrowser.cdp.exceptions import CDPError


if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Generator, MutableMapping

    from trio_websocket import WebSocketConnection
    from typing_extensions import Self

    from streamlink.webbrowser.cdp.devtools.target import TargetID
    from streamlink.webbrowser.cdp.devtools.util import T_JSON_DICT


log = getLogger(__name__)

MAX_BUFFER_SIZE = 10
MAX_MESSAGE_SIZE = 2**24  # ~16MiB
CMD_TIMEOUT = 2

TCmdResponse = TypeVar("TCmdResponse")
TCDPEvent = TypeVar("TCDPEvent", bound=CDPEvent)


class CDPEventListener(Generic[TCDPEvent]):
    """
    Instances of this class are returned by :meth:`CDPBase.listen()`.

    The return types of each of its methods depend on the event type.

    Can be used as an async for-loop which indefinitely waits for events to be emitted,
    or can be used as an async context manager which yields the event and closes the listener when leaving the context manager.

    Example:

    .. code-block:: python

        async def listen(cdp_session: CDPSession):
            async for request in cdp_session.listen(devtools.fetch.RequestPaused):
                ...

        async def listen_once(cdp_session: CDPSession):
            async with cdp_session.listen(devtools.fetch.RequestPaused) as request:
                ...
    """

    _sender: trio.MemorySendChannel[TCDPEvent]
    _receiver: trio.MemoryReceiveChannel[TCDPEvent]

    def __init__(
        self,
        event_channels: MutableMapping[type[TCDPEvent], set[trio.MemorySendChannel[TCDPEvent]]],
        event: type[TCDPEvent],
        max_buffer_size: int | None = None,
    ):
        max_buffer_size = MAX_BUFFER_SIZE if max_buffer_size is None else max_buffer_size
        self._sender, self._receiver = trio.open_memory_channel(max_buffer_size)
        event_channels[event].add(self._sender)

    async def receive(self) -> TCDPEvent:
        """
        Await a single event without closing the listener's memory channel.
        """

        return await self._receiver.receive()

    def close(self) -> None:
        self._receiver.close()

    async def __aenter__(self) -> TCDPEvent:
        return await self._receiver.receive()

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        # sync
        self.close()

    def __aiter__(self) -> Self:
        return self

    async def __anext__(self) -> TCDPEvent:
        try:
            return await self._receiver.receive()
        except trio.EndOfChannel as err:
            self.close()
            raise StopAsyncIteration from err

    def __del__(self) -> None:
        self.close()


@dataclasses.dataclass
class _CDPCmdBuffer(Generic[TCmdResponse]):
    cmd: Generator[dict, dict, TCmdResponse]
    response: TCmdResponse | Exception | None = None
    event: trio.Event = dataclasses.field(default_factory=trio.Event)

    def set_response(self, response: TCmdResponse | Exception) -> None:
        self.response = response
        self.event.set()


# The design of CDPBase/CDPConnection/CDPSession is based on the trio-chrome-devtools-protocol project version 0.6.0
# https://github.com/HyperionGray/trio-chrome-devtools-protocol/blob/0.6.0/trio_cdp/__init__.py
#
# The MIT License (MIT)
#
# Copyright (c) 2018 Hyperion Gray
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


class CDPBase:
    """
    Low-level base class for Chrome Devtools Protocol connection & session management.

    It provides methods for sending CDP commands and receiving their responses, as well as for listening to CDP events.

    Both CDP commands and events can be sent and received in a global context and in a session context.

    The Chrome Devtools Protocol is documented at https://chromedevtools.github.io/devtools-protocol/
    """

    def __init__(
        self,
        websocket: WebSocketConnection,
        target_id: TargetID | None = None,
        session_id: SessionID | None = None,
        cmd_timeout: float = CMD_TIMEOUT,
    ) -> None:
        self.websocket = websocket
        self.target_id = target_id
        self.session_id = session_id
        self.cmd_timeout = cmd_timeout
        self.event_channels: MutableMapping[type[CDPEvent], set[trio.MemorySendChannel[CDPEvent]]] = defaultdict(set)
        self.cmd_buffers: dict[int, _CDPCmdBuffer] = {}
        self.cmd_id = itertools.count()

    async def send(
        self,
        cmd: Generator[T_JSON_DICT, T_JSON_DICT, TCmdResponse],
        timeout: float | None = None,
    ) -> TCmdResponse:
        """
        Send a specific CDP command and await its response.

        :param cmd: See the ``streamlink.webbrowser.cdp.devtools`` package for the available commands.
        :param timeout: Override of the max amount of time a response can take. Uses the class's default value otherwise.
                        This override is mostly only relevant for awaiting JS code evaluations.
        :return: The return value depends on the used command.
        """

        cmd_id = next(self.cmd_id)
        cmd_buffer = _CDPCmdBuffer(cmd)

        self.cmd_buffers[cmd_id] = cmd_buffer

        cmd_data = next(cmd)
        cmd_data["id"] = cmd_id
        if self.session_id:
            cmd_data["sessionId"] = self.session_id

        message = json.dumps(cmd_data, separators=(",", ":"), sort_keys=True)
        log.all("Sending message: %(message)s", dict(message=message))
        with trio.move_on_after(self.cmd_timeout if timeout is None else timeout) as cancel_scope:
            try:
                await self.websocket.send_message(message)
            except ConnectionClosed as err:
                self.cmd_buffers.pop(cmd_id, None)
                raise CDPError(err.reason) from err

            await cmd_buffer.event.wait()
        if cancel_scope.cancel_called:
            self.cmd_buffers.pop(cmd_id, None)
            raise CDPError("Sending CDP message and receiving its response timed out")

        response = cast("TCmdResponse", cmd_buffer.response)
        self.cmd_buffers.pop(cmd_id, None)

        if isinstance(response, Exception):
            raise response

        return response

    def listen(self, event: type[TCDPEvent], max_buffer_size: int | None = None) -> CDPEventListener[TCDPEvent]:
        """
        Listen to a CDP event and return a new :class:`CDPEventListener` instance.

        :param event: See the ``streamlink.webbrowser.cdp.devtools`` package for the available events.
                      For events to be sent over the CDP connection, a specific domain needs to be enabled first.
        :param max_buffer_size: The buffer size of the ``trio`` memory channel.
        :return:
        """

        return CDPEventListener(self.event_channels, event, max_buffer_size)  # type: ignore

    def _handle_data(self, data: T_JSON_DICT) -> None:
        if "id" in data:
            self._handle_cmd_response(data)
        else:
            self._handle_event(data)

    def _handle_cmd_response(self, data: T_JSON_DICT) -> None:
        try:
            cmd_id: int = data["id"]
            cmd_buffer = self.cmd_buffers.pop(cmd_id)
        except KeyError:
            log.warning("Got a CDP command response with an unknown ID: %(id)r", dict(id=data.get("id")))
            return

        if "error" in data:
            cmd_buffer.set_response(CDPError(f"Error in CDP command response {cmd_id}: {data['error']}"))
            return
        if "result" not in data:
            cmd_buffer.set_response(CDPError(f"No result in CDP command response {cmd_id}"))
            return

        cmd_result: T_JSON_DICT = data["result"]
        try:
            # send the response to the command's generator function (the first send() must stop the generator)
            cmd_buffer.cmd.send(cmd_result)
        except StopIteration as cm:
            # and on success, set the response result
            cmd_buffer.set_response(cm.value)
        except Exception as err:
            # handle any errors raised by the generator's result logic
            cmd_buffer.set_response(CDPError(f"Generator of CDP command ID {cmd_id} raised {type(err).__name__}: {err}"))
        else:
            cmd_buffer.set_response(CDPError(f"Generator of CDP command ID {cmd_id} did not exit when expected!"))

    def _handle_event(self, data: T_JSON_DICT) -> None:
        if "method" not in data or "params" not in data:
            log.warning("Invalid CDP event message received without method or params")
            return

        try:
            event = parse_json_event(data)
        except KeyError:
            log.warning(f"Unknown CDP event message received: {data['method']}")
            return

        log.all("Received event: %(event)r", dict(event=event))
        broken_channels = set()
        for sender in self.event_channels[type(event)]:
            try:
                sender.send_nowait(event)
            except trio.WouldBlock:
                log.error("Unable to propagate CDP event %(event)r due to full channel", dict(event=event))
            except trio.BrokenResourceError:
                broken_channels.add(sender)
                sender.close()
        self.event_channels[type(event)] -= broken_channels


class CDPConnection(CDPBase, trio.abc.AsyncResource):
    """
    Don't instantiate this class yourself, see its :meth:`create()` classmethod.
    """

    def __init__(self, websocket: WebSocketConnection, cmd_timeout: float) -> None:
        super().__init__(websocket=websocket, cmd_timeout=cmd_timeout)
        self.sessions: dict[SessionID, CDPSession] = {}

    @classmethod
    @asynccontextmanager
    async def create(cls, url: str, timeout: float | None = None) -> AsyncGenerator[Self, None]:
        """
        Establish a new CDP connection to the Chromium-based web browser's remote debugging interface.

        :param url: The websocket address
        :param timeout: The max amount of time a single CDP command response can take.
        :return:
        """

        async with trio.open_nursery() as nursery:
            websocket = await connect_websocket_url(nursery, url, max_message_size=MAX_MESSAGE_SIZE)
            cdp_connection = cls(websocket, timeout or CMD_TIMEOUT)
            nursery.start_soon(cdp_connection._task_reader)
            try:
                yield cdp_connection
            finally:
                await cdp_connection.aclose()

    async def aclose(self) -> None:
        """
        Close the websocket connection, close all memory channels and clean up all opened sessions.
        """

        await self.websocket.aclose()
        inst: CDPBase
        for inst in (self, *self.sessions.values()):
            for event_channels in inst.event_channels.values():
                for event_channel in event_channels:
                    event_channel.close()
            inst.event_channels.clear()
        self.sessions.clear()

    async def new_target(self, url: str = "") -> CDPSession:
        """
        Create a new target (browser tab) and return a new :class:`CDPSession` instance.

        :param url: Optional URL. Leave empty for a blank target (preferred for proper navigation handling).
        :return:
        """

        target_id = await self.send(create_target(url))

        return await self.get_session(target_id)

    async def get_session(self, target_id: TargetID) -> CDPSession:
        session_id = await self.send(attach_to_target(target_id, True))
        cdp_session = CDPSession(self.websocket, target_id=target_id, session_id=session_id, cmd_timeout=self.cmd_timeout)
        self.sessions[session_id] = cdp_session

        return cdp_session

    async def _task_reader(self) -> None:
        while True:
            try:
                message = await self.websocket.get_message()
            except ConnectionClosed:
                break

            try:
                data = json.loads(message)
            except json.JSONDecodeError as err:
                raise CDPError(f"Received invalid CDP JSON data: {err}") from err

            log.all("Received message: %(message)s", dict(message=message))
            if "sessionId" not in data:
                self._handle_data(data)
            else:
                session_id = SessionID(data["sessionId"])
                if session_id not in self.sessions:
                    raise CDPError(f"Unknown CDP session ID: {session_id!r}")
                self.sessions[session_id]._handle_data(data)


class CDPSession(CDPBase):
    pass



"""
$description Indian live TV channels and video on-demand service. OTT service from YuppTV.
$url yupptv.com
$type live, vod
$account Some streams require an account and subscription
"""

import re
import time

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginargument, pluginmatcher
from streamlink.plugin.api import useragents
from streamlink.stream.hls import HLSStream, HLSStreamReader, HLSStreamWriter


log = getLogger(__name__)


class HLSStreamWriterYupptv(HLSStreamWriter):
    def should_filter_segment(self, segment):
        return ".ts" not in segment.uri or super().should_filter_segment(segment)


class HLSStreamReaderYupptv(HLSStreamReader):
    __writer__ = HLSStreamWriterYupptv


class HLSStreamYupptv(HLSStream):
    __reader__ = HLSStreamReaderYupptv


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?yupptv\.com"),
)
@pluginargument(
    "boxid",
    requires=["yuppflixtoken"],
    sensitive=True,
    metavar="BOXID",
    help="The yupptv.com boxid that's used in the BoxId cookie.",
)
@pluginargument(
    "yuppflixtoken",
    sensitive=True,
    metavar="YUPPFLIXTOKEN",
    help="The yupptv.com yuppflixtoken that's used in the YuppflixToken cookie.",
)
@pluginargument(
    "purge-credentials",
    action="store_true",
    help="Purge cached YuppTV credentials to initiate a new session and reauthenticate.",
)
class YuppTV(Plugin):
    _cookie_expiry = 3600 * 24 * 365

    def _login_using_box_id_and_yuppflix_token(self, box_id, yuppflix_token):
        time_now = time.time()

        self.session.http.cookies.set(
            "BoxId",
            box_id,
            domain="www.yupptv.com",
            path="/",
            expires=time_now + self._cookie_expiry,
        )
        self.session.http.cookies.set(
            "YuppflixToken",
            yuppflix_token,
            domain="www.yupptv.com",
            path="/",
            expires=time_now + self._cookie_expiry,
        )

        self.save_cookies()
        log.info("Successfully set BoxId and YuppflixToken")

    def _get_streams(self):
        self.session.http.headers.update({"User-Agent": useragents.CHROME})
        self.session.http.headers.update({"Origin": "https://www.yupptv.com"})

        authed = (
            self.session.http.cookies.get("BoxId")
            and self.session.http.cookies.get("YuppflixToken")
        )  # fmt: skip

        login_box_id = self.get_option("boxid")
        login_yuppflix_token = self.get_option("yuppflixtoken")

        if self.options.get("purge_credentials"):
            self.clear_cookies()
            authed = False
            log.info("All credentials were successfully removed")

        if authed:
            log.debug("Attempting to authenticate using cached cookies")
        elif login_box_id and login_yuppflix_token:
            self._login_using_box_id_and_yuppflix_token(
                login_box_id,
                login_yuppflix_token,
            )
            authed = True

        page = self.session.http.get(self.url)
        if authed and "btnsignup" in page.text:
            log.error("This device requires renewed credentials to log in")
            return

        match = re.search(r"""src:\s*(?P<q>["'])(?P<url>.+?)(?P=q)""", page.text)
        if not match or "preview/" in match["url"]:
            if "btnsignup" in page.text:
                log.error("This stream requires you to login")
            else:
                log.error("This stream requires a subscription")
            return

        def override_encoding(res, *_, **__):
            res.encoding = "utf-8"

        return HLSStreamYupptv.parse_variant_playlist(
            self.session,
            match["url"],
            hooks={"response": override_encoding},
        )


__plugin__ = YuppTV



import io
from threading import Thread

from streamlink.buffers import Buffer, RingBuffer


class StreamIOWrapper(io.IOBase):
    """Wraps file-like objects that are not inheriting from IOBase"""

    def __init__(self, fd):
        self.fd = fd

    def read(self, size=-1):
        return self.fd.read(size)

    def close(self):
        if hasattr(self.fd, "close"):
            self.fd.close()


class StreamIOIterWrapper(io.IOBase):
    """Wraps a iterator and turn it into a file-like object"""

    def __init__(self, iterator):
        self.iterator = iterator
        self.buffer = Buffer()

    def read(self, size=-1):
        if size < 0:
            size = self.buffer.length

        while self.buffer.length < size:
            try:
                chunk = next(self.iterator)
                self.buffer.write(chunk)
            except StopIteration:
                break

        return self.buffer.read(size)

    def close(self):
        pass


class StreamIOThreadWrapper(io.IOBase):
    """Wraps a file-like object in a thread.

    Useful for getting control over read timeout where
    timeout handling is missing or out of our control.
    """

    class Filler(Thread):
        def __init__(self, fd, buffer):
            super().__init__()

            self.error = None
            self.fd = fd
            self.buffer = buffer
            self.daemon = True
            self.running = False

        def run(self):
            self.running = True

            while self.running:
                try:
                    data = self.fd.read(8192)
                except OSError as error:
                    self.error = error
                    break

                if len(data) == 0:
                    break

                self.buffer.write(data)

            self.stop()

        def stop(self):
            self.running = False
            self.buffer.close()

            if hasattr(self.fd, "close"):
                try:
                    self.fd.close()
                except Exception:
                    pass

    def __init__(self, session, fd, timeout=30):
        self.buffer = RingBuffer(session.get_option("ringbuffer-size"))
        self.fd = fd
        self.timeout = timeout

        self.filler = StreamIOThreadWrapper.Filler(self.fd, self.buffer)
        self.filler.start()

    def read(self, size=-1):
        if self.filler.error and self.buffer.length == 0:
            raise self.filler.error

        return self.buffer.read(size, block=self.filler.is_alive(), timeout=self.timeout)

    def close(self):
        self.filler.stop()

        if self.filler.is_alive():
            self.filler.join()


__all__ = ["StreamIOWrapper", "StreamIOIterWrapper", "StreamIOThreadWrapper"]



from __future__ import annotations

import base64
import hashlib
import importlib.metadata
import json
import pkgutil
import re
from contextlib import suppress
from pathlib import Path
from typing import TYPE_CHECKING, Literal, TypeAlias, TypedDict, cast

import streamlink.plugins
from streamlink.logger import getLogger
from streamlink.options import Argument, Arguments

# noinspection PyProtectedMember
from streamlink.plugin.plugin import _PLUGINARGUMENT_TYPE_REGISTRY, NO_PRIORITY, NORMAL_PRIORITY, Matcher, Matchers, Plugin
from streamlink.utils.module import exec_module, get_finder


if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping
    from types import ModuleType

    from _typeshed.importlib import PathEntryFinderProtocol


log = getLogger(".".join(__name__.split(".")[:-1]))

# The path to Streamlink's built-in plugins
_PLUGINS_PATH = Path(streamlink.plugins.__path__[0])

# Hardcoded plugins JSON file path
_PLUGINSDATA_PATH = _PLUGINS_PATH / "_plugins.json"
# Hardcoded package name to look up metadata
_PLUGINSDATA_PACKAGENAME = "streamlink"
# The `parts` value of the plugins JSON file contained in the package's `RECORD` metadata file
_PLUGINSDATA_PACKAGEPATH = "streamlink", "plugins", "_plugins.json"


class StreamlinkPlugins:
    """
    Streamlink's session-plugins implementation. This class is responsible for loading plugins and resolving them from URLs.

    See the :attr:`Streamlink.plugins <streamlink.session.Streamlink.plugins>` attribute.

    Unless disabled by the user, Streamlink will try to load built-in plugins lazily, when accessing them for the first time
    while resolving input URLs. This is done by reading and interpreting serialized data of each plugin's
    :func:`pluginmatcher <streamlink.plugin.pluginmatcher>` and :func:`pluginargument <streamlink.plugin.pluginargument>`
    data from a pre-built plugins JSON file which is included in Streamlink's wheel packages.

    Plugins which are sideloaded, either from specific user directories or custom input directories,
    always have a higher priority than built-in plugins.
    """

    def __init__(self, builtin: bool = True, lazy: bool = True):
        # Loaded plugins
        self._plugins: dict[str, type[Plugin]] = {}

        # Data of built-in plugins which can be loaded lazily
        self._matchers: dict[str, Matchers] = {}
        self._arguments: dict[str, Arguments] = {}

        # Attempt to load built-in plugins lazily first
        if builtin and lazy:
            data = StreamlinkPluginsData.load()
            if data:
                self._matchers, self._arguments = data
            else:
                lazy = False

        # Load built-ins if lazy-loading is disabled or if loading plugins data has failed
        if builtin and not lazy:
            self.load_builtin()

    def __getitem__(self, item: str) -> type[Plugin]:
        """Access a loaded plugin class by name"""
        return self._plugins[item]

    def __setitem__(self, key: str, value: type[Plugin]) -> None:
        """Add/override a plugin class by name"""
        self._plugins[key] = value

    def __delitem__(self, key: str) -> None:
        """Remove a loaded plugin by name"""
        self._plugins.pop(key, None)

    def __contains__(self, item: str) -> bool:
        """Check if a plugin is loaded"""
        return item in self._plugins

    def get_names(self) -> list[str]:
        """Get a list of the names of all available plugins"""
        return sorted(self._plugins.keys() | self._matchers.keys())

    def get_loaded(self) -> dict[str, type[Plugin]]:
        """Get a mapping of all loaded plugins"""
        return dict(self._plugins)

    def load_builtin(self) -> bool:
        """Load Streamlink's built-in plugins"""
        return self.load_path(_PLUGINS_PATH)

    def load_path(self, path: str | Path) -> bool:
        """Load plugins from a custom directory"""
        plugins = self._load_plugins_from_path(path)
        self.update(plugins)

        return bool(plugins)

    def update(self, plugins: Mapping[str, type[Plugin]]):
        """Add/override loaded plugins"""
        self._plugins.update(plugins)

    def clear(self):
        """Remove all loaded plugins from the session"""
        self._plugins.clear()

    def iter_arguments(self) -> Iterator[tuple[str, Arguments]]:
        """Iterate through all plugins and their :class:`Arguments <streamlink.options.Arguments>`"""
        yield from (
            (name, plugin.arguments)
            for name, plugin in self._plugins.items()
            if plugin.arguments
        )  # fmt: skip
        yield from (
            (name, arguments)
            for name, arguments in self._arguments.items()
            if arguments and name not in self._plugins
        )  # fmt: skip

    def iter_matchers(self) -> Iterator[tuple[str, Matchers]]:
        """Iterate through all plugins and their :class:`Matchers <streamlink.plugin.plugin.Matchers>`"""
        yield from (
            (name, plugin.matchers)
            for name, plugin in self._plugins.items()
            if plugin.matchers
        )  # fmt: skip
        yield from (
            (name, matchers)
            for name, matchers in self._matchers.items()
            if matchers and name not in self._plugins
        )  # fmt: skip

    def match_url(self, url: str) -> tuple[str, type[Plugin]] | None:
        """Find a matching plugin by URL and load plugins which haven't been loaded yet"""
        match: str | None = None
        priority: int = NO_PRIORITY

        for name, matchers in self.iter_matchers():
            for matcher in matchers:
                if matcher.priority > priority and matcher.pattern.match(url) is not None:
                    match = name
                    priority = matcher.priority

        if match is None:
            return None

        # plugin not loaded yet?
        # if a custom plugin with the same name has already been loaded, skip loading the built-in plugin
        if match not in self._plugins:
            log.debug(f"Loading plugin: {match}")
            lookup = self._load_plugin_from_path(match, _PLUGINS_PATH)
            if not lookup:
                return None
            self._plugins[match] = lookup[1]

        return match, self._plugins[match]

    def _load_plugin_from_path(self, name: str, path: Path) -> tuple[ModuleType, type[Plugin]] | None:
        finder = get_finder(path)

        return self._load_plugin_from_finder(name, finder)

    def _load_plugins_from_path(self, path: str | Path) -> dict[str, type[Plugin]]:
        plugins: dict[str, type[Plugin]] = {}
        for finder, name, _ in pkgutil.iter_modules([str(path)]):
            finder = cast("PathEntryFinderProtocol", finder)
            lookup = self._load_plugin_from_finder(name, finder=finder)
            if lookup is None:
                continue
            mod, plugin = lookup
            if (name in self._plugins or name in self._matchers) and mod.__file__:
                with Path(mod.__file__).open("rb") as fh:
                    sha256 = hashlib.sha256(fh.read())
                log.info(f"Plugin {name} is being overridden by {mod.__file__} (sha256:{sha256.hexdigest()})")
            plugins[name] = plugin

        return plugins

    @staticmethod
    def _load_plugin_from_finder(name: str, finder: PathEntryFinderProtocol) -> tuple[ModuleType, type[Plugin]] | None:
        try:
            # set the full plugin module name, even for sideloaded plugins
            mod = exec_module(finder, f"streamlink.plugins.{name}", override=True)
        except ImportError as err:
            log.exception(f"Failed to load plugin {name} from {err.path}\n")
            return None

        if not hasattr(mod, "__plugin__") or not issubclass(mod.__plugin__, Plugin):
            return None

        return mod, mod.__plugin__


_RE_STRIP_JSON_COMMENTS = re.compile(rb"^(?:\s*//[^\n]*\n+)+")

if TYPE_CHECKING:
    _TListOfConstants: TypeAlias = list[bool | int | float | str | None]
    _TConstantOrListOfConstants: TypeAlias = bool | int | float | str | _TListOfConstants | None
    _TMappingOfConstantOrListOfConstants: TypeAlias = dict[str, _TConstantOrListOfConstants]


class _TPluginMatcherData(TypedDict):
    pattern: str
    flags: int | None
    priority: int | None
    name: str | None


class _TPluginArgumentData(TypedDict):
    name: str
    action: str | None
    nargs: int | Literal["*", "?", "+"] | None
    const: _TConstantOrListOfConstants
    default: _TConstantOrListOfConstants
    type: str | None
    type_args: _TListOfConstants | None
    type_kwargs: _TMappingOfConstantOrListOfConstants | None
    choices: _TListOfConstants | None
    required: bool | None
    help: str | None
    metavar: str | list[str] | None
    dest: str | None
    requires: str | list[str] | None
    prompt: str | None
    sensitive: bool | None
    argument_name: str | None


class _TPluginData(TypedDict):
    matchers: list[_TPluginMatcherData]
    arguments: list[_TPluginArgumentData]


class StreamlinkPluginsData:
    @classmethod
    def load(cls) -> tuple[dict[str, Matchers], dict[str, Arguments]] | None:
        # specific errors get logged, others are ignored intentionally
        with suppress(Exception):
            content = _PLUGINSDATA_PATH.read_bytes()

            cls._validate(content)

            return cls._parse(content)

        return None

    @staticmethod
    def _validate(content: bytes) -> None:
        # find plugins JSON checksum in package metadata
        # https://packaging.python.org/en/latest/specifications/recording-installed-packages/#the-record-file
        mode, filehash = next(
            (packagepath.hash.mode, packagepath.hash.value)
            for packagepath in importlib.metadata.files(_PLUGINSDATA_PACKAGENAME) or []
            if packagepath.hash is not None and packagepath.parts == _PLUGINSDATA_PACKAGEPATH
        )
        if mode not in hashlib.algorithms_guaranteed or not filehash:
            log.error("Unknown plugins data hash mode, falling back to loading all plugins")
            raise Exception

        # compare checksums
        hashalg = getattr(hashlib, mode)
        hashobj = hashalg(content)
        digest = base64.urlsafe_b64encode(hashobj.digest()).decode("utf-8").rstrip("=")
        if digest != filehash:
            log.error("Plugins data checksum mismatch, falling back to loading all plugins")
            raise Exception

    @classmethod
    def _parse(cls, content: bytes) -> tuple[dict[str, Matchers], dict[str, Arguments]]:
        content = _RE_STRIP_JSON_COMMENTS.sub(b"", content)
        data: dict[str, _TPluginData] = json.loads(content)

        try:
            matchers = cls._build_matchers(data)
        except Exception:
            log.exception("Error while loading pluginmatcher data from JSON")
            raise

        try:
            arguments = cls._build_arguments(data)
        except Exception:
            log.exception("Error while loading pluginargument data from JSON")
            raise

        return matchers, arguments

    @classmethod
    def _build_matchers(cls, data: dict[str, _TPluginData]) -> dict[str, Matchers]:
        res = {}
        for pluginname, plugindata in data.items():
            matchers = Matchers()
            for m in plugindata.get("matchers") or []:
                matcher = cls._build_matcher(m)
                matchers.add(matcher)

            res[pluginname] = matchers

        return res

    @staticmethod
    def _build_matcher(data: _TPluginMatcherData) -> Matcher:
        return Matcher(
            pattern=re.compile(data.get("pattern"), data.get("flags") or 0),
            priority=data.get("priority") or NORMAL_PRIORITY,
            name=data.get("name"),
        )

    @classmethod
    def _build_arguments(cls, data: dict[str, _TPluginData]) -> dict[str, Arguments]:
        res = {}
        for pluginname, plugindata in data.items():
            if not plugindata.get("arguments"):
                continue
            arguments = Arguments()
            for a in reversed(plugindata.get("arguments") or []):
                if argument := cls._build_argument(a):
                    arguments.add(argument)

            res[pluginname] = arguments

        return res

    @staticmethod
    def _build_argument(data: _TPluginArgumentData) -> Argument | None:
        name = data.get("name")
        type_data = data.get("type")
        if not type_data:
            argument_type = None
        elif argument_type := _PLUGINARGUMENT_TYPE_REGISTRY.get(type_data):
            type_args = data.get("type_args") or ()
            type_kwargs = data.get("type_kwargs") or {}
            if type_args or type_kwargs:
                argument_type = argument_type(*type_args, **type_kwargs)
        else:
            return None

        return Argument(
            name=name,
            action=data.get("action"),
            nargs=data.get("nargs"),
            const=data.get("const"),
            default=data.get("default"),
            type=argument_type,
            choices=data.get("choices"),
            required=data.get("required") or False,
            help=data.get("help"),
            metavar=data.get("metavar"),
            dest=data.get("dest"),
            requires=data.get("requires"),
            prompt=data.get("prompt"),
            sensitive=data.get("sensitive") or False,
            argument_name=data.get("argument_name"),
        )



"""
$description Vietnamese live TV channels from THVL, including THVL1, THVL2, THVL3 and THVL4.
$url thvli.vn
$type live
$metadata id
$metadata title
$region Vietnam
"""

import re
from datetime import datetime, timezone
from hashlib import md5

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?thvli\.vn/live/(?P<channel>[^/]+)"),
)
class VinhLongTV(Plugin):
    _API_URL = "https://api.thvli.vn/backend/cm/get_detail/{channel}/"
    _API_KEY_DATE = "Kh0ngDuLieu"
    _API_KEY_TIME = "C0R0i"
    _API_KEY_SECRET = "Kh0aAnT0an"

    def _get_headers(self):
        now = datetime.now(tz=timezone.utc)
        date = now.strftime("%Y%m%d")
        time = now.strftime("%H%M%S")
        dtstr = f"{date}{time}"
        dthash = md5(dtstr.encode()).hexdigest()
        key_value = f"{dthash[:3]}{dthash[-3:]}"
        key_access = f"{self._API_KEY_DATE}{date}{self._API_KEY_TIME}{time}{self._API_KEY_SECRET}{key_value}"

        return {
            "X-SFD-Date": dtstr,
            "X-SFD-Key": md5(key_access.encode()).hexdigest(),
        }

    def _get_streams(self):
        channel = self.match.group("channel")
        params = {"timezone": "UTC"}
        headers = self._get_headers()

        self.id, self.title, hls_url = self.session.http.get(
            self._API_URL.format(channel=channel),
            params=params,
            headers=headers,
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "id": str,
                    "title": str,
                    "link_play": str,
                },
                validate.union_get(
                    "id",
                    "title",
                    "link_play",
                ),
            ),
        )

        return HLSStream.parse_variant_playlist(self.session, hls_url)


__plugin__ = VinhLongTV



"""
$description Live TV channel and video on-demand service from TV5Monde, a French free-to-air broadcaster.
$url tv5monde.com
$url tivi5mondeplus.com
$type live, vod
$region France, Belgium, Switzerland, Africa
"""

import re
from urllib.parse import quote

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.dash import DASHStream
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream
from streamlink.utils.url import update_scheme


@pluginmatcher(
    re.compile(r"https?://(?:[\w-]+\.)*(?:tv5monde|tivi5mondeplus)\.com/"),
)
class TV5Monde(Plugin):
    _URL_API = "https://api.tv5monde.com/player/asset/{asset}/resolve"

    _MAP_STREAMTYPES = {
        "application/x-mpegURL": HLSStream.parse_variant_playlist,
        "application/dash+xml": DASHStream.parse_manifest,
        "video/mp4": lambda *args: {"vod": HTTPStream(*args)},
    }

    def _find_streams(self, data):
        schema_streams = validate.Schema([
            validate.all(
                {
                    "type": str,
                    "url": validate.url(),
                },
                validate.union_get("type", "url"),
            ),
        ])
        streams = schema_streams.validate(data)

        for mimetype, streamfactory in self._MAP_STREAMTYPES.items():
            for streamtype, streamurl in streams:
                if streamtype == mimetype:
                    url = update_scheme("https://", streamurl)
                    return streamfactory(self.session, url)

    def _get_playlist(self, data):
        schema_playlist = validate.Schema(
            validate.parse_json(),
            [{"sources": list}],
            validate.get((0, "sources")),
        )
        playlist = schema_playlist.validate(data)

        return self._find_streams(playlist)

    def _get_broadcast(self, data):
        schema_broadcast = validate.Schema(
            validate.parse_json(),
            [
                {
                    "type": str,
                    "url": str,
                },
            ],
        )
        broadcast = schema_broadcast.validate(data)

        deferred = next((item["url"] for item in broadcast if item["type"] == "application/deferred"), None)
        if deferred:
            url = self._URL_API.format(asset=quote(deferred))
            broadcast = self.session.http.get(
                url,
                schema=validate.Schema(
                    validate.parse_json(),
                ),
            )

        return self._find_streams(broadcast)

    def _get_streams(self):
        playlist, broadcast = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.union((
                    validate.xml_xpath_string(".//*[@data-playlist][1]/@data-playlist"),
                    validate.xml_xpath_string(".//*[@data-broadcast][1]/@data-broadcast"),
                )),
            ),
        )

        # data-playlist needs higher priority (data-broadcast attribute exists on the same DOM node)
        if playlist:
            return self._get_playlist(playlist)
        if broadcast:
            return self._get_broadcast(broadcast)


__plugin__ = TV5Monde



from __future__ import annotations

import socket
from contextlib import suppress
from http.server import BaseHTTPRequestHandler
from io import BytesIO

from streamlink_cli.output.abc import Output


class HTTPRequest(BaseHTTPRequestHandler):
    # noinspection PyMissingConstructor
    def __init__(self, request_text):
        self.rfile = BytesIO(request_text)
        self.raw_requestline = self.rfile.readline()
        self.error_code = self.error_message = None
        self.parse_request()

    def send_error(self, code, message=None, explain=None):
        self.error_code = code
        self.error_message = message


class HTTPOutput(Output):
    socket: socket.socket

    def __init__(self, host: str | None = "127.0.0.1", port: int = 0) -> None:
        super().__init__()
        self.host = host
        self.port = port
        self.conn: socket.socket | None = None

    @property
    def addresses(self) -> list[str]:
        if self.host:
            return [self.host]

        addrs = {"127.0.0.1"}
        with suppress(socket.gaierror):
            addrinfo = socket.getaddrinfo(socket.gethostname(), self.port, socket.AF_INET)
            for _family, _type, _proto, _canonname, (address, *_) in addrinfo:
                if not isinstance(address, str):
                    continue
                addrs.add(address)

        return sorted(addrs)

    @property
    def urls(self):
        for addr in self.addresses:
            yield f"http://{addr}:{self.port}/"

    @property
    def url(self):
        return next(self.urls, None)

    def start_server(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind((self.host or "", self.port))
        self.socket.listen(1)
        self.host, self.port = self.socket.getsockname()
        if self.host == "0.0.0.0":
            self.host = None

    def accept_connection(self, timeout=30) -> None:
        self.socket.settimeout(timeout)

        try:
            conn, _addr = self.socket.accept()
            conn.settimeout(None)
            self.conn = conn
        except TimeoutError as err:
            self.conn = None
            raise OSError("Socket accept timed out") from err

    def _open(self):
        conn = self.conn
        if not conn:
            raise OSError("No client connection")

        try:
            req_data = conn.recv(1024)
        except OSError as err:
            raise OSError("Failed to read data from socket") from err

        req = HTTPRequest(req_data)
        if req.command not in ("GET", "HEAD"):
            conn.send(b"HTTP/1.1 501 Not Implemented\r\n")
            conn.close()
            raise OSError(f"Invalid request method: {req.command}")

        try:
            conn.send(b"HTTP/1.1 200 OK\r\n")
            conn.send(b"Server: Streamlink\r\n")
            conn.send(b"Content-Type: video/unknown\r\n")
            conn.send(b"\r\n")
        except OSError as err:
            raise OSError("Failed to write data to socket") from err

        # We don't want to send any data on HEAD requests.
        if req.command == "HEAD":
            conn.close()
            raise OSError

        self.request = req

    def _write(self, data):
        self.conn.sendall(data)  # type: ignore

    def _close(self):
        if self.conn:
            with suppress(OSError):
                self.conn.close()
            self.conn = None

    def shutdown(self) -> None:
        self.close()
        with suppress(OSError):
            self.socket.shutdown(socket.SHUT_RDWR)
        with suppress(OSError):
            self.socket.close()



ANDROID = "Mozilla/5.0 (Linux; Android 16) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/147.0.7727.114 Mobile Safari/537.36"
CHROME = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/147.0.0.0 Safari/537.36"
CHROME_OS = "Mozilla/5.0 (X11; CrOS x86_64 16610.44.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/147.0.7727.115 Safari/537.36"
FIREFOX = "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:150.0) Gecko/20100101 Firefox/150.0"
IE_11 = "Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; rv:11.0) like Gecko"
IPHONE = "Mozilla/5.0 (iPhone; CPU iPhone OS 18_7_8 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/26.0 Mobile/15E148 Safari/604.1"
OPERA = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/147.0.0.0 Safari/537.36 OPR/131.0.0.0"
SAFARI = "Mozilla/5.0 (Macintosh; Intel Mac OS X 15_7_5) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/26.0 Safari/605.1.15"

ANDROID_VERSION = (16,)
CHROME_VERSION = (147, 0, 7727, 139)
CHROME_OS_VERSION = (147, 0, 7727, 115)
FIREFOX_VERSION = (150, 0, 1)
IE_11_VERSION = (11,)
IPHONE_VERSION = (18, 7, 8)
OPERA_VERSION = (131, 0, 5877, 5)
SAFARI_VERSION = (26,)

# Backwards compatibility
EDGE = CHROME
FIREFOX_MAC = FIREFOX
IE_6 = IE_7 = IE_8 = IE_9 = IE_11
IPHONE_6 = IPAD = IPHONE
SAFARI_7 = SAFARI_8 = SAFARI
WINDOWS_PHONE_8 = ANDROID

DEFAULT = FIREFOX



"""
$description Social platform delivering live broadcasts on diverse topics, from politics and music to entertainment.
$url 17app.co
$type live
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://17\.live/.+/live/(?P<channel>[^/&?]+)"),
)
class App17(Plugin):
    def _get_streams(self):
        channel = self.match.group("channel")
        self.session.http.headers.update({"Referer": self.url})
        data = self.session.http.post(
            f"https://wap-api.17app.co/api/v1/lives/{channel}/viewers/alive",
            data={"liveStreamID": channel},
            schema=validate.Schema(
                validate.parse_json(),
                validate.any(
                    {
                        "rtmpUrls": [
                            {
                                validate.optional("provider"): validate.any(int, None),
                                "url": validate.url(path=validate.endswith(".flv")),
                            },
                        ],
                    },
                    {"errorCode": int, "errorMessage": str},
                ),
            ),
            acceptable_status=(200, 403, 404, 420),
        )
        log.trace("%r", data)
        if data.get("errorCode"):
            log.error(f"{data['errorCode']} - {data['errorMessage'].replace('Something wrong: ', '')}")
            return

        flv_url = data["rtmpUrls"][0]["url"]
        yield "live", HTTPStream(self.session, flv_url)

        if "wansu-" in flv_url:
            hls_url = flv_url.replace(".flv", "/playlist.m3u8")
        else:
            hls_url = flv_url.replace("live-hdl", "live-hls").replace(".flv", ".m3u8")

        s = HLSStream.parse_variant_playlist(self.session, hls_url)
        if not s:
            yield "live", HLSStream(self.session, hls_url)
        elif len(s) == 1:
            yield "live", next(iter(s.values()))
        else:
            yield from s.items()


__plugin__ = App17



"""
$description Global video game live-streaming and video hosting platform, owned by Tencent.
$url trovo.live
$type live, vod
$metadata id
$metadata author
$metadata category
$metadata title
"""

import re
import secrets
import sys

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.utils.url import update_scheme


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?trovo\.live/s/(?P<user>[^/?&]+)(?:/\d+\?vid=(?P<video_id>[^/?&]+))?"),
)
class Trovo(Plugin):
    @classmethod
    def stream_weight(cls, stream: str) -> tuple[float, str]:
        if stream == "source":
            return sys.maxsize, stream
        return super().stream_weight(stream)

    @staticmethod
    def generate_qid():
        return f"{secrets.token_hex(8)}".upper()

    def get_vod(self, video_id):
        json = self.session.http.post(
            f"https://api-web.trovo.live/graphql?qid={self.generate_qid()}",
            json=[
                {
                    "operationName": "vod_VodReaderService_BatchGetVodDetailInfo",
                    "variables": {
                        "params": {
                            "vids": [video_id],
                        },
                    },
                    "extensions": {},
                },
            ],
            schema=validate.Schema(
                validate.parse_json(),
                [
                    {
                        "data": {
                            "vod_VodReaderService_BatchGetVodDetailInfo": {
                                "VodDetailInfos": validate.any(
                                    {
                                        video_id: {
                                            "streamerInfo": {
                                                "userName": str,
                                            },
                                            "vodInfo": {
                                                "playInfos": [
                                                    {
                                                        "desc": validate.all(validate.transform(lambda s: s.lower()), str),
                                                        "playUrl": validate.url(),
                                                    },
                                                ],
                                                "vid": str,
                                                "title": str,
                                                "categoryName": str,
                                                "playbackRights": {
                                                    "playbackRightsSetting": str,
                                                    "playbackRights": str,
                                                },
                                            },
                                        },
                                    },
                                    {},
                                ),
                            },
                        },
                    },
                ],
                validate.get((0, "data", "vod_VodReaderService_BatchGetVodDetailInfo", "VodDetailInfos", video_id)),
            ),
        )

        if not json:
            log.error("Video not found")
            return

        log.debug(json["vodInfo"]["playbackRights"])
        self.id = json["vodInfo"]["vid"]
        self.author = json["streamerInfo"]["userName"]
        self.title = json["vodInfo"]["title"]
        self.category = json["vodInfo"]["categoryName"]

        for s in json["vodInfo"]["playInfos"]:
            q = s["desc"]
            if "(source)" in q:
                q = "source"
            yield q, HLSStream(self.session, update_scheme("https:", s["playUrl"]))

    def get_live(self, user):
        json = self.session.http.post(
            f"https://api-web.trovo.live/graphql?qid={self.generate_qid()}",
            json=[
                {
                    "operationName": "live_LiveReaderService_GetLiveInfo",
                    "variables": {
                        "params": {
                            "userName": user,
                        },
                    },
                },
            ],
            schema=validate.Schema(
                validate.parse_json(),
                validate.any(
                    [
                        {
                            "data": {
                                validate.optional("live_LiveReaderService_GetLiveInfo"): {
                                    "streamerInfo": {
                                        "userName": str,
                                    },
                                    "programInfo": {
                                        "id": str,
                                        "title": str,
                                        "streamInfo": [
                                            {
                                                "desc": validate.all(validate.transform(lambda s: s.lower()), str),
                                                "playUrl": validate.transform(lambda s: s.replace(".flv?", ".m3u8?")),
                                            },
                                        ],
                                    },
                                    "categoryInfo": {
                                        "shortName": str,
                                    },
                                    "isLive": int,
                                },
                            },
                        },
                    ],
                    [
                        {
                            "errors": [
                                {
                                    "message": validate.transform(lambda s: s.replace('\\"', '"')),
                                },
                            ],
                        },
                    ],
                ),
                validate.get(0),
            ),
        )

        if "errors" in json:
            errors = [e["message"] for e in json["errors"]]
            log.error(f"API error(s): {', '.join(errors)}")
            return

        if json["data"]["live_LiveReaderService_GetLiveInfo"]:
            json = json["data"]["live_LiveReaderService_GetLiveInfo"]
        else:
            log.error("Stream data not found")
            return

        if not json["isLive"]:
            log.error("This stream is no longer live")
            return

        self.id = json["programInfo"]["id"]
        self.author = json["streamerInfo"]["userName"]
        self.title = json["programInfo"]["title"]
        self.category = json["categoryInfo"]["shortName"]

        for s in json["programInfo"]["streamInfo"]:
            if s["playUrl"]:
                yield s["desc"], HLSStream(self.session, update_scheme("https:", s["playUrl"]))

    def _get_streams(self):
        self.session.http.headers.update({"Origin": "https://trovo.live"})
        url_data = self.match.groupdict()

        if url_data["video_id"]:
            return self.get_vod(url_data["video_id"])
        elif url_data["user"]:
            return self.get_live(url_data["user"])


__plugin__ = Trovo



"""
$description Global live-streaming and video on-demand platform owned by IBM.
$url ustream.tv
$url video.ibm.com
$type live, vod
"""

from __future__ import annotations

import re
from collections import deque
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from random import randint
from threading import Event, RLock
from typing import TYPE_CHECKING, Any
from urllib.parse import urljoin, urlunparse

from requests import Response

from streamlink.exceptions import PluginError, StreamError
from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginargument, pluginmatcher
from streamlink.plugin.api import useragents, validate
from streamlink.plugin.api.websocket import WebsocketClient
from streamlink.stream.ffmpegmux import MuxedStream
from streamlink.stream.segmented import Segment, SegmentedStreamReader, SegmentedStreamWorker, SegmentedStreamWriter
from streamlink.stream.stream import Stream
from streamlink.utils.parse import parse_json


if TYPE_CHECKING:
    from collections.abc import Callable, Mapping


log = getLogger(__name__)


@dataclass
class _StreamFormat:
    contentType: str
    sourceStreamVersion: int
    initUrl: str
    segmentUrl: str
    bitrate: int


@dataclass
class StreamFormatVideo(_StreamFormat):
    height: int


@dataclass
class StreamFormatAudio(_StreamFormat):
    language: str = ""


@dataclass
class UStreamTVSegment(Segment):
    available_at: datetime
    hash: str
    path: str

    # the segment URLs depend on the CDN and the chosen stream format and its segment template string
    def url(self, base: str | None, template: str) -> str:
        return urljoin(
            base or "",
            f"{self.path}/{template.replace('%', str(self.num), 1).replace('%', self.hash, 1)}",
        )


class UStreamTVWsClient(WebsocketClient):
    API_URL = "wss://r{0}-1-{1}-{2}-ws-{3}.ums.services.video.ibm.com/1/ustream"
    APP_ID = 3
    APP_VERSION = 2

    STREAM_OPENED_TIMEOUT = 6

    _schema_cmd = validate.Schema({
        "cmd": str,
        "args": [{str: object}],
    })
    _schema_stream_formats = validate.Schema({
        "streams": [
            validate.any(
                validate.all(
                    {
                        "contentType": "video/mp4",
                        "sourceStreamVersion": int,
                        "initUrl": str,
                        "segmentUrl": str,
                        "bitrate": int,
                        "height": int,
                    },
                    validate.transform(lambda obj: StreamFormatVideo(**obj)),
                ),
                validate.all(
                    {
                        "contentType": "audio/mp4",
                        "sourceStreamVersion": int,
                        "initUrl": str,
                        "segmentUrl": str,
                        "bitrate": int,
                        validate.optional("language"): str,
                    },
                    validate.transform(lambda obj: StreamFormatAudio(**obj)),
                ),
                object,
            ),
        ],
    })
    _schema_stream_segments = validate.Schema({
        "chunkId": int,
        "chunkTime": int,
        "contentAccess": validate.all(
            {
                "accessList": [
                    {
                        "data": {
                            "path": str,
                        },
                    },
                ],
            },
            validate.get(("accessList", 0, "data", "path")),
        ),
        "hashes": {validate.transform(int): str},
    })

    stream_cdn: str | None = None
    stream_formats_video: list[StreamFormatVideo] | None = None
    stream_formats_audio: list[StreamFormatAudio] | None = None
    stream_initial_id: int | None = None

    def __init__(
        self,
        session,
        media_id,
        application,
        referrer=None,
        cluster="live",
        password=None,
        app_id=APP_ID,
        app_version=APP_VERSION,
    ) -> None:
        self.opened = Event()
        self.ready = Event()
        self.stream_error = None
        # a list of deques subscribed by worker threads which independently need to read segments
        self.stream_segments_subscribers: list[deque[UStreamTVSegment]] = []
        self.stream_segments_initial: deque[UStreamTVSegment] = deque()
        self.stream_segments_lock = RLock()

        self.media_id = media_id
        self.application = application
        self.referrer = referrer
        self.cluster = cluster
        self.password = password
        self.app_id = app_id
        self.app_version = app_version

        super().__init__(session, self._get_url(), origin="https://www.ustream.tv")

    def _get_url(self):
        return self.API_URL.format(randint(0, 0xFFFFFF), self.media_id, self.application, self.cluster)

    def _set_error(self, error: Any):
        self.stream_error = error
        self.ready.set()

    def _set_ready(self):
        if not self.ready.is_set() and self.stream_cdn and self.stream_initial_id is not None:
            self.ready.set()

            if self.opened.wait(self.STREAM_OPENED_TIMEOUT):
                log.debug("Stream opened, keeping websocket connection alive")
            else:
                log.info("Closing websocket connection")
                self.ws.close()

    def segments_subscribe(self) -> deque[UStreamTVSegment]:
        with self.stream_segments_lock:
            # copy the initial segments deque (segments arrive early)
            new_deque = self.stream_segments_initial.copy()
            self.stream_segments_subscribers.append(new_deque)

            return new_deque

    def _segments_append(self, segment: UStreamTVSegment):
        # if there are no subscribers yet, add segment(s) to the initial deque
        if not self.stream_segments_subscribers:
            self.stream_segments_initial.append(segment)
        else:
            for subscriber_deque in self.stream_segments_subscribers:
                subscriber_deque.append(segment)

    def on_open(self, wsapp):
        args = {
            "type": "viewer",
            "appId": self.app_id,
            "appVersion": self.app_version,
            "rsid": f"{randint(0, 10_000_000_000):x}:{randint(0, 10_000_000_000):x}",
            "rpin": f"_rpin.{randint(0, 1_000_000_000_000_000)}",
            "referrer": self.referrer,
            "clusterHost": "r%rnd%-1-%mediaId%-%mediaType%-%protocolPrefix%-%cluster%.ums.ustream.tv",
            "media": self.media_id,
            "application": self.application,
        }
        if self.password:
            args["password"] = self.password

        self.send_json({
            "cmd": "connect",
            "args": [args],
        })

    def on_message(self, wsapp, data: str):
        try:
            parsed = parse_json(data, schema=self._schema_cmd)
        except PluginError:
            log.error(f"Could not parse message: {data[:50]}")
            return

        cmd: str = parsed["cmd"]
        args: list[dict] = parsed["args"]
        log.trace("Received '%s' command", cmd)
        log.trace("%r", args)

        handlers = self._MESSAGE_HANDLERS.get(cmd)
        if handlers is not None:
            for arg in args:
                for name, handler in handlers.items():
                    argdata = arg.get(name)
                    if argdata is not None:
                        log.debug(f"Processing '{cmd}' - '{name}'")
                        handler(self, argdata)

    # noinspection PyMethodMayBeStatic
    def _handle_warning(self, data: dict):
        log.warning(f"{data['code']}: {str(data['message'])[:50]}")

    # noinspection PyUnusedLocal
    def _handle_reject_nonexistent(self, *args):
        self._set_error("This channel does not exist")

    # noinspection PyUnusedLocal
    def _handle_reject_geo_lock(self, *args):
        self._set_error("This content is not available in your area")

    def _handle_reject_cluster(self, arg: dict):
        self.cluster = arg["name"]
        log.info(f"Switching cluster to: {self.cluster}")
        self.reconnect(url=self._get_url())

    def _handle_reject_referrer_lock(self, arg: dict):
        self.referrer = arg["redirectUrl"]
        log.info(f"Updating referrer to: {self.referrer}")
        self.reconnect(url=self._get_url())

    def _handle_module_info_cdn_config(self, data: dict):
        self.stream_cdn = urlunparse((
            data["protocol"],
            data["data"][0]["data"][0]["sites"][0]["host"],
            data["data"][0]["data"][0]["sites"][0]["path"],
            "",
            "",
            "",
        ))
        self._set_ready()

    def _handle_module_info_stream(self, data: dict):
        if data.get("contentAvailable") is False:
            return self._set_error("This stream is currently offline")

        mp4_segmented = data.get("streamFormats", {}).get("mp4/segmented")
        if not mp4_segmented:
            return

        # parse the stream formats once
        if self.stream_initial_id is None:
            try:
                formats = self._schema_stream_formats.validate(mp4_segmented)
                formats = formats["streams"]
            except PluginError as err:
                return self._set_error(err)
            self.stream_formats_video = list(filter(lambda f: type(f) is StreamFormatVideo, formats))
            self.stream_formats_audio = list(filter(lambda f: type(f) is StreamFormatAudio, formats))

        # parse segment duration and hashes, and queue new segments
        try:
            segmentdata: dict = self._schema_stream_segments.validate(mp4_segmented)
        except PluginError:
            log.error("Failed parsing hashes")
            return

        current_id: int = segmentdata["chunkId"]
        duration: int = segmentdata["chunkTime"]
        path: str = segmentdata["contentAccess"]
        hashes: dict[int, str] = segmentdata["hashes"]

        sorted_ids = sorted(hashes.keys())
        count = len(sorted_ids)
        if count == 0:
            return

        # initial segment ID (needed by the workers to filter queued segments)
        if self.stream_initial_id is None:
            self.stream_initial_id = current_id

        current_time = datetime.now(timezone.utc)

        # lock the stream segments deques for the worker threads
        with self.stream_segments_lock:
            # interpolate and extrapolate segments from the provided id->hash data
            diff = 10 - sorted_ids[0] % 10  # if there's only one id->hash item, extrapolate until the next decimal
            for idx, segment_id in enumerate(sorted_ids):
                idx_next = idx + 1
                if idx_next < count:
                    # calculate the difference between IDs and use that to interpolate segment IDs
                    # the last id->hash item will use the previous diff to extrapolate segment IDs
                    diff = sorted_ids[idx_next] - segment_id
                for num in range(segment_id, segment_id + diff):
                    self._segments_append(
                        UStreamTVSegment(
                            uri="",
                            num=num,
                            duration=duration,
                            available_at=current_time + timedelta(seconds=(num - current_id - 1) * duration / 1000),
                            hash=hashes[segment_id],
                            path=path,
                        ),
                    )

        self._set_ready()

    # ----

    _MESSAGE_HANDLERS: Mapping[str, Mapping[str, Callable[[UStreamTVWsClient, Any], None]]] = {
        "warning": {
            "code": _handle_warning,
        },
        "reject": {
            "cluster": _handle_reject_cluster,
            "referrerLock": _handle_reject_referrer_lock,
            "nonexistent": _handle_reject_nonexistent,
            "geoLock": _handle_reject_geo_lock,
        },
        "moduleInfo": {
            "cdnConfig": _handle_module_info_cdn_config,
            "stream": _handle_module_info_stream,
        },
    }


class UStreamTVStreamWriter(SegmentedStreamWriter[UStreamTVSegment, Response]):
    reader: UStreamTVStreamReader
    stream: UStreamTVStream

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._has_init = False

    def put(self, segment):
        if self.closed:  # pragma: no cover
            return

        if segment is None:
            self.queue(None, None)
        else:
            if not self._has_init:
                self._has_init = True
                self.queue(segment, self.executor.submit(self.fetch, segment, True))
            self.queue(segment, self.executor.submit(self.fetch, segment, False))

    # noinspection PyMethodOverriding
    def fetch(self, segment: UStreamTVSegment, is_init: bool):  # type: ignore[override, ty:invalid-method-override]
        if self.closed:  # pragma: no cover
            return

        now = datetime.now(timezone.utc)
        if segment.available_at > now:
            time_to_wait = (segment.available_at - now).total_seconds()
            log.debug(f"Waiting for {self.stream.kind} segment: {segment.num} ({time_to_wait:.01f}s)")
            if not self.reader.worker.wait(time_to_wait):
                return

        try:
            return self.session.http.get(
                segment.url(
                    self.stream.wsclient.stream_cdn,
                    self.stream.stream_format.initUrl if is_init else self.stream.stream_format.segmentUrl,
                ),
                timeout=self.timeout,
                retries=self.retries,
                exception=StreamError,
            )
        except StreamError as err:
            log.error(f"Failed to fetch {self.stream.kind} segment {segment.num}: {err}")

    def write(self, segment: UStreamTVSegment, result: Response, *data):
        if self.closed:  # pragma: no cover
            return
        try:
            for chunk in result.iter_content(8192):
                self.reader.buffer.write(chunk)
            log.debug(f"Download of {self.stream.kind} segment {segment.num} complete")
        except OSError as err:
            log.error(f"Failed to read {self.stream.kind} segment {segment.num}: {err}")


class UStreamTVStreamWorker(SegmentedStreamWorker[UStreamTVSegment, Response]):
    reader: UStreamTVStreamReader
    writer: UStreamTVStreamWriter
    stream: UStreamTVStream

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.wsclient = self.stream.wsclient
        self.sequence = self.wsclient.stream_initial_id or -1
        self.queue = self.wsclient.segments_subscribe()

    def iter_segments(self):
        duration = 5000
        while not self.closed:
            try:
                with self.wsclient.stream_segments_lock:
                    segment = self.queue.popleft()
                    duration = segment.duration
            except IndexError:
                # wait for new segments to be queued (half the last segment's duration in seconds)
                if self.wait(duration / 1000 / 2):
                    continue

            if self.closed:
                return

            if segment.num < self.sequence:
                continue

            log.debug(f"Adding {self.stream.kind} segment {segment.num} to queue")
            yield segment
            self.sequence = segment.num + 1


class UStreamTVStreamReader(SegmentedStreamReader[UStreamTVSegment, Response]):
    __worker__ = UStreamTVStreamWorker
    __writer__ = UStreamTVStreamWriter

    stream: UStreamTVStream
    worker: UStreamTVStreamWorker
    writer: UStreamTVStreamWriter

    def open(self):
        self.stream.wsclient.opened.set()
        super().open()

    def close(self):
        super().close()
        self.stream.wsclient.close()


class UStreamTVStream(Stream):
    __shortname__ = "ustreamtv"

    def __init__(
        self,
        session,
        kind: str,
        wsclient: UStreamTVWsClient,
        stream_format: StreamFormatVideo | StreamFormatAudio,
    ):
        super().__init__(session)
        self.kind = kind
        self.wsclient = wsclient
        self.stream_format = stream_format

    def open(self):
        reader = UStreamTVStreamReader(self)
        reader.open()

        return reader


@pluginmatcher(
    re.compile(
        r"""
            https?://(?:(?:www\.)?ustream\.tv|video\.ibm\.com)
            (?:
                /combined-embed
                /(?P<combined_channel_id>\d+)
                (?:/video/(?P<combined_video_id>\d+))?
                |
                (?:(?:/embed/|/channel/(?:id/)?)(?P<channel_id>\d+))?
                (?:(?:/embed)?/recorded/(?P<video_id>\d+))?
            )
        """,
        re.VERBOSE,
    ),
)
@pluginargument(
    "password",
    sensitive=True,
    argument_name="ustream-password",
    metavar="PASSWORD",
    help="A password to access password protected UStream.tv channels.",
)
class UStreamTV(Plugin):
    STREAM_READY_TIMEOUT = 15

    def _get_media_app(self):
        video_id = self.match.group("video_id") or self.match.group("combined_video_id")
        if video_id:
            return video_id, "recorded"

        channel_id = self.match.group("channel_id") or self.match.group("combined_channel_id")
        if not channel_id:
            channel_id = self.session.http.get(
                self.url,
                headers={"User-Agent": useragents.CHROME},
                schema=validate.Schema(
                    validate.parse_html(),
                    validate.xml_xpath_string(".//meta[@name='ustream:channel_id'][@content][1]/@content"),
                ),
            )

        return channel_id, "channel"

    def _get_streams(self):
        if not MuxedStream.is_usable(self.session):
            return

        media_id, application = self._get_media_app()
        if not media_id:
            return

        wsclient = UStreamTVWsClient(
            self.session,
            media_id,
            application,
            referrer=self.url,
            cluster="live",
            password=self.get_option("password"),
        )
        log.debug(
            "Connecting to UStream API: "
            + ", ".join([
                f"media_id={media_id}",
                f"application={application}",
                f"referrer={self.url}",
                f"cluster={'live'}",
            ]),
        )
        wsclient.start()

        log.debug(f"Waiting for stream data (for at most {self.STREAM_READY_TIMEOUT} seconds)...")
        if (
            not wsclient.ready.wait(self.STREAM_READY_TIMEOUT)
            or not wsclient.is_alive()
            or wsclient.stream_error
        ):  # fmt: skip
            log.error(wsclient.stream_error or "Waiting for stream data timed out.")
            wsclient.close()
            return

        if not wsclient.stream_formats_audio:
            for video in wsclient.stream_formats_video or []:
                yield f"{video.height}p", UStreamTVStream(self.session, "video", wsclient, video)
        else:
            for video in wsclient.stream_formats_video or []:
                for audio in wsclient.stream_formats_audio:
                    yield (
                        f"{video.height}p+a{audio.bitrate}k",
                        MuxedStream(
                            self.session,
                            UStreamTVStream(self.session, "video", wsclient, video),
                            UStreamTVStream(self.session, "audio", wsclient, audio),
                        ),
                    )


__plugin__ = UStreamTV



from collections import deque
from io import BytesIO
from threading import Event, Lock


class Chunk(BytesIO):
    """A single chunk, part of the buffer."""

    def __init__(self, buf):
        self.length = len(buf)
        BytesIO.__init__(self, buf)

    @property
    def empty(self):
        return self.tell() == self.length


class Buffer:
    """Simple buffer for use in single-threaded consumer/filler.

    Stores chunks in a deque to avoid inefficient reallocating
    of large buffers.
    """

    def __init__(self):
        self.chunks = deque()
        self.current_chunk = None
        self.closed = False
        self.length = 0
        self.written_once = False

    def _iterate_chunks(self, size):
        bytes_left = size

        while bytes_left:
            try:
                current_chunk = self.current_chunk or Chunk(self.chunks.popleft())
            except IndexError:
                break

            data = current_chunk.read(bytes_left)
            bytes_left -= len(data)

            if current_chunk.empty:
                self.current_chunk = None
            else:
                self.current_chunk = current_chunk

            yield data

    def write(self, data):
        if not self.closed:
            data = bytes(data)  # Copy so that original buffer may be reused
            self.chunks.append(data)
            self.length += len(data)
            self.written_once = True

    def read(self, size=-1):
        if size < 0 or size > self.length:
            size = self.length

        if not size:
            return b""

        data = b"".join(self._iterate_chunks(size))
        self.length -= len(data)

        return data

    def close(self):
        self.closed = True


class RingBuffer(Buffer):
    """Circular buffer for use in multi-threaded consumer/filler."""

    def __init__(self, size=8192 * 4):
        Buffer.__init__(self)

        self.buffer_size = size
        self.buffer_lock = Lock()

        self.event_free = Event()
        self.event_free.set()
        self.event_used = Event()

    def _check_events(self):
        if self.is_full:
            self.event_free.clear()
        else:
            self.event_free.set()

        if self.length > 0:
            self.event_used.set()
        else:
            self.event_used.clear()

    def _read(self, size=-1):
        with self.buffer_lock:
            data = Buffer.read(self, size)

            self._check_events()

        return data

    def read(self, size=-1, block=True, timeout=None):
        if block and not self.closed:
            if not self.event_used.wait(timeout) and self.length == 0:
                raise OSError("Read timeout")

        return self._read(size)

    def write(self, data):
        if self.closed:
            return

        data_left = len(data)
        data_total = len(data)

        while data_left > 0:
            self.event_free.wait()

            if self.closed:
                return

            with self.buffer_lock:
                write_len = min(self.free, data_left)
                written = data_total - data_left

                Buffer.write(self, data[written : written + write_len])
                data_left -= write_len

                self._check_events()

    def resize(self, size):
        with self.buffer_lock:
            self.buffer_size = size

            self._check_events()

    def wait_free(self, timeout=None):
        return self.event_free.wait(timeout)

    def wait_used(self, timeout=None):
        return self.event_used.wait(timeout)

    def close(self):
        Buffer.close(self)

        # Make sure we don't let a .write() and .read() block forever
        self.event_free.set()
        self.event_used.set()

    @property
    def free(self):
        return max(self.buffer_size - self.length, 0)

    @property
    def is_full(self):
        return self.free == 0


__all__ = ["Buffer", "RingBuffer"]



from streamlink.exceptions import PluginError
from streamlink.options import Argument as PluginArgument, Arguments as PluginArguments, Options as PluginOptions
from streamlink.plugin.plugin import (
    HIGH_PRIORITY,
    LOW_PRIORITY,
    NO_PRIORITY,
    NORMAL_PRIORITY,
    Plugin,
    pluginargument,
    pluginmatcher,
)


__all__ = [
    "HIGH_PRIORITY",
    "NORMAL_PRIORITY",
    "LOW_PRIORITY",
    "NO_PRIORITY",
    "Plugin",
    "PluginArguments",
    "PluginArgument",
    "PluginError",
    "PluginOptions",
    "pluginmatcher",
    "pluginargument",
]



# ruff: noqa: RUF067

from signal import SIGINT, SIGTERM, signal
from sys import exit  # noqa: A004


def _exit(*_):
    # don't raise a KeyboardInterrupt until streamlink_cli has been fully initialized
    exit(128 | 2)


# override default SIGINT handler (and set SIGTERM handler) as early as possible
signal(SIGINT, _exit)
signal(SIGTERM, _exit)



"""
$description French live TV channels from TF1 Group, including LCI and TF1.
$url tf1.fr
$url tf1info.fr
$url lci.fr
$type live
$region France
$account Required on tf1.fr
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, PluginError, pluginargument, pluginmatcher
from streamlink.plugin.api import useragents, validate
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    name="live",
    pattern=re.compile(r"https?://(?:www\.)?tf1\.fr/(?P<live>(?![\w-]+-\d+)[^/?#]+)/direct/?"),
)
@pluginmatcher(
    name="stream",
    pattern=re.compile(r"https?://(?:www\.)?tf1\.fr/(?P<stream>[\w-]+-\d+)/direct/?"),
)
@pluginmatcher(
    name="lci",
    pattern=re.compile(r"https?://(?:www\.)?(?:tf1info|lci)\.fr/direct/?"),
)
@pluginargument(
    "email",
    requires=["password"],
    metavar="EMAIL",
    help="The email address used to register with tf1.fr.",
)
@pluginargument(
    "password",
    sensitive=True,
    metavar="PASSWORD",
    help="A tf1.fr account password to use with --tf1-email.",
)
@pluginargument(
    "purge-credentials",
    action="store_true",
    help="Purge cached tf1.fr credentials to initiate a new session and reauthenticate.",
)
class TF1(Plugin):
    _URL_API = "https://mediainfo.tf1.fr/mediainfocombo/{channel_id}"
    _URL_LOGIN = "https://compte.tf1.fr/accounts.login"
    _URL_TOKEN = "https://www.tf1.fr/token/gigya/web"
    _API_KEY = "3_hWgJdARhz_7l1oOp3a8BDLoR9cuWZpUaKG4aqF7gum9_iK3uTZ2VlDBl8ANf8FVk"
    # Gigya GDPR consent IDs
    _CONSENT_IDS = ["4", "10001", "10003", "10005", "10007", "10009", "10011", "10013", "10015", "10017", "10019"]

    _CACHE_KEY_USER_TOKEN = "token"

    def _login(self, login_id, password):
        status, *login_data = self.session.http.post(
            url=self._URL_LOGIN,
            data={
                "loginID": login_id,
                "password": password,
                "APIKey": self._API_KEY,
                "includeUserInfo": "true",
            },
            schema=validate.Schema(
                validate.parse_json(),
                validate.any(
                    validate.all(
                        {
                            "errorCode": 0,
                            "UID": str,
                            "UIDSignature": str,
                            "signatureTimestamp": str,
                        },
                        validate.union_get("UID", "UIDSignature", "signatureTimestamp"),
                        validate.transform(lambda data: ("success", *data)),
                    ),
                    validate.all(
                        {
                            "errorCode": int,
                            "errorDetails": str,
                        },
                        validate.union_get("errorCode", "errorDetails"),
                        validate.transform(lambda data: ("failure", *data)),
                    ),
                ),
            ),
        )

        if status != "success":
            error_code, error_details = login_data
            raise PluginError(f"{error_code=} - {error_details or 'Unknown error'}")

        uid, uid_signature, signature_timestamp = login_data
        log.debug(f"{uid=} {uid_signature=} {signature_timestamp=}")

        return self.session.http.post(
            url=self._URL_TOKEN,
            json={
                "uid": uid,
                "signature": uid_signature,
                "timestamp": int(signature_timestamp),
                "consent_ids": self._CONSENT_IDS,
            },
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "token": str,
                },
                validate.get("token"),
            ),
        )

    def _get_channel(self):
        if self.matches["live"]:
            channel = self.match["live"]
            channel_id = f"L_{channel.upper()}"
        elif self.matches["stream"]:
            channel = self.match["stream"]
            channel_id = f"L_FAST_v2l-{channel}"
        elif self.matches["lci"]:
            channel = "LCI"
            channel_id = "L_LCI"
        else:  # pragma: no cover
            raise PluginError("Invalid channel")

        return channel, channel_id

    def _api_call(self, channel_id, user_token):
        headers = {
            # forces HLS streams
            "User-Agent": useragents.IPHONE,
        }
        if user_token:
            headers["Authorization"] = f"Bearer {user_token}"

        return self.session.http.get(
            self._URL_API.format(channel_id=channel_id),
            params={
                "context": "MYTF1",
                "pver": "5015000",
            },
            headers=headers,
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "delivery": validate.any(
                        validate.all(
                            {
                                "code": 200,
                                "format": "hls",
                                "url": validate.url(),
                            },
                            validate.union_get("code", "url"),
                        ),
                        validate.all(
                            {
                                "code": int,
                                "error": str,
                            },
                            validate.union_get("code", "error"),
                        ),
                    ),
                },
                validate.get("delivery"),
            ),
        )

    def _get_streams(self):
        channel, channel_id = self._get_channel()
        log.debug(f"Found channel {channel} ({channel_id})")

        if self.get_option("purge-credentials"):
            log.info("Removing cached user-authentication token...")
            self.cache.set(self._CACHE_KEY_USER_TOKEN, None, 0)
            user_token = None
        else:
            user_token = self.cache.get(self._CACHE_KEY_USER_TOKEN)

        if (
            not user_token
            and (login_email := self.get_option("email"))
            and (login_password := self.get_option("password"))
        ):  # fmt: skip
            log.info("Acquiring new user-authentication token...")
            user_token = self._login(login_email, login_password)
            self.cache.set(self._CACHE_KEY_USER_TOKEN, user_token)

        code, data = self._api_call(channel_id, user_token)
        if code != 200:
            log.error(data)
            return

        return HLSStream.parse_variant_playlist(self.session, data)


__plugin__ = TF1



"""
$description Japanese live-streaming service used primarily by Japanese idols & voice actors and their fans.
$url showroom-live.com
$type live
$metadata title
"""

import re
from urllib.parse import parse_qsl, urlparse

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:\w+\.)?showroom-live\.com/"),
)
class Showroom(Plugin):
    LIVE_STATUS = 2

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.session.set_option("hls-playlist-reload-time", "segment")

    def _get_streams(self):
        room_id = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//nav//a[contains(@href,'/room/profile?')]/@href"),
                validate.none_or_all(
                    validate.transform(lambda _url_profile: dict(parse_qsl(urlparse(_url_profile).query))),
                    validate.get("room_id"),
                ),
            ),
        )
        if not room_id:
            return

        log.debug(f"Room ID: {room_id}")

        live_status, self.title = self.session.http.get(
            "https://www.showroom-live.com/api/live/live_info",
            params={
                "room_id": room_id,
            },
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "live_status": int,
                    "room_name": str,
                },
                validate.union_get(
                    "live_status",
                    "room_name",
                ),
            ),
        )
        if live_status != self.LIVE_STATUS:
            log.info("This stream is currently offline")
            return

        url = self.session.http.get(
            "https://www.showroom-live.com/api/live/streaming_url",
            params={
                "room_id": room_id,
                "abr_available": 1,
            },
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "streaming_url_list": [
                        {
                            "type": str,
                            "url": validate.url(),
                        },
                    ],
                },
                validate.get("streaming_url_list"),
                validate.filter(lambda p: p["type"] == "hls_all"),
                validate.get((0, "url")),
            ),
        )

        res = self.session.http.get(url, acceptable_status=(200, 403, 404))
        if res.headers["Content-Type"] not in ("application/x-mpegURL", "application/vnd.apple.mpegurl"):
            log.error("This stream is restricted")
            return

        return HLSStream.parse_variant_playlist(self.session, url)


__plugin__ = Showroom



"""
$description American live TV channel and video on-demand service owned by Warner Bros.
$url adultswim.com
$type live, vod
$region various
$notes VODs may be protected by DRM
"""

import re
from urllib.parse import urlparse, urlunparse

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, PluginError, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.utils.parse import parse_json


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?adultswim\.com/(streams|videos)(?:/([^/]+))?(?:/([^/]+))?"),
)
class AdultSwim(Plugin):
    token_url = "https://token.ngtv.io/token/token_spe"
    video_data_url = "https://www.adultswim.com/api/shows/v1/media/{0}/desktop"

    app_id_js_url_re = re.compile(
        r"""<script src="([^"]*asvp\..*?\.bundle\.js)">""",
    )

    app_id_re = re.compile(
        r'''CDN_TOKEN_APP_ID="(.*?)"''',
    )

    json_data_re = re.compile(
        r"""<script id="__NEXT_DATA__" type="application/json">({.*})</script>""",
    )

    truncate_url_re = re.compile(r"""(.*)/\w+/?""")

    _api_schema = validate.Schema(
        {
            "media": {
                "desktop": {
                    str: {
                        "url": validate.url(),
                    },
                },
            },
        },
        validate.get("media"),
        validate.get("desktop"),
        validate.filter(lambda k, v: k in ["unprotected", "bulkaes"]),
    )

    _stream_data_schema = validate.Schema(
        {
            "props": {
                "__REDUX_STATE__": {
                    "streams": [
                        {
                            "id": str,
                            "stream": str,
                        },
                    ],
                },
            },
        },
        validate.get("props"),
        validate.get("__REDUX_STATE__"),
        validate.get("streams"),
    )

    _token_schema = validate.Schema(
        validate.any(
            {"auth": {"token": str}},
            {"auth": {"error": {"message": str}}},
        ),
        validate.get("auth"),
    )

    _video_data_schema = validate.Schema(
        {
            "props": {
                "pageProps": {
                    "__APOLLO_STATE__": {
                        str: {
                            validate.optional("id"): str,
                            validate.optional("slug"): str,
                        },
                    },
                },
            },
        },
        validate.get("props"),
        validate.get("pageProps"),
        validate.get("__APOLLO_STATE__"),
        validate.filter(lambda k, v: k.startswith("Video:")),
    )

    def _get_stream_data(self, streamid):
        res = self.session.http.get(self.url)
        m = self.json_data_re.search(res.text)
        if m and m.group(1):
            streams = parse_json(m.group(1), schema=self._stream_data_schema)
        else:
            raise PluginError("Failed to get json_data")

        for stream in streams:
            if "id" in stream and streamid == stream["id"] and "stream" in stream:
                return stream["stream"]

    def _get_video_data(self, slug):
        m = self.truncate_url_re.search(self.url)
        if m and m.group(1):
            log.debug(f"Truncated URL={m.group(1)}")
        else:
            raise PluginError("Failed to truncate URL")

        res = self.session.http.get(m.group(1))
        m = self.json_data_re.search(res.text)
        if m and m.group(1):
            videos = parse_json(m.group(1), schema=self._video_data_schema)
        else:
            raise PluginError("Failed to get json_data")

        for video in videos:
            if "slug" in videos[video]:
                if slug == videos[video]["slug"] and "id" in videos[video]:
                    return videos[video]["id"]

    def _get_token(self, path):
        res = self.session.http.get(self.url)
        m = self.app_id_js_url_re.search(res.text)
        app_id_js_url = m and m.group(1)
        if not app_id_js_url:
            raise PluginError("Could not determine app_id_js_url")
        log.debug(f"app_id_js_url={app_id_js_url}")

        res = self.session.http.get(app_id_js_url)
        m = self.app_id_re.search(res.text)
        app_id = m and m.group(1)
        if not app_id:
            raise PluginError("Could not determine app_id")
        log.debug(f"app_id={app_id}")

        res = self.session.http.get(
            self.token_url,
            params=dict(
                format="json",
                appId=app_id,
                path=path,
            ),
        )

        token_data = self.session.http.json(res, schema=self._token_schema)
        if "error" in token_data:
            raise PluginError(token_data["error"]["message"])

        return token_data["token"]

    def _get_streams(self):
        url_type, show_name, episode_name = self.match.groups()

        if url_type == "streams" and not show_name:
            url_type = "live-stream"
        elif not show_name:
            raise PluginError(f"Missing show_name for url_type: {url_type}")

        log.debug(f"URL type={url_type}")

        if url_type == "live-stream":
            video_id = self._get_stream_data(url_type)
        elif url_type == "streams":
            video_id = self._get_stream_data(show_name)
        elif url_type == "videos":
            if show_name is None or episode_name is None:
                raise PluginError(
                    f"Missing show_name or episode_name for url_type: {url_type}",
                )
            video_id = self._get_video_data(episode_name)
        else:
            raise PluginError(f"Unrecognised url_type: {url_type}")

        if video_id is None:
            raise PluginError("Could not find video_id")
        log.debug(f"Video ID={video_id}")

        res = self.session.http.get(self.video_data_url.format(video_id))

        url_data = self.session.http.json(res, schema=self._api_schema)
        if "unprotected" in url_data:
            url = url_data["unprotected"]["url"]
        elif "bulkaes" in url_data:
            url_parsed = urlparse(url_data["bulkaes"]["url"])
            token = self._get_token(url_parsed.path)
            url = urlunparse((
                url_parsed.scheme,
                url_parsed.netloc,
                url_parsed.path,
                url_parsed.params,
                f"hdnts={token}",
                url_parsed.fragment,
            ))
        else:
            raise PluginError("Could not find a usable URL in url_data")

        log.debug(f"URL={url}")

        return HLSStream.parse_variant_playlist(self.session, url)


__plugin__ = AdultSwim



"""
$description Live TV channels and video on-demand service from RUV, an Icelandic public, state-owned broadcaster.
$url ruv.is
$type live, vod
$region Iceland
"""

import re
from textwrap import dedent

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream


log = getLogger(__name__)


@pluginmatcher(
    name="live",
    pattern=re.compile(r"https?://(?:www\.)?ruv\.is/(?:sjonvarp|utvarp)/beint/(?P<channel>\w+)$"),
)
@pluginmatcher(
    name="vod",
    pattern=re.compile(r"https?://(?:www\.)?ruv\.is/(?:sjonvarp|utvarp)/spila/[^/]+/(?P<id>\d+)/(?P<episode>[^/]+)/?"),
)
class Ruv(Plugin):
    _URL_API_CHANNEL = "https://geo.spilari.ruv.is/channel/{channel}"
    _URL_API_GQL = "https://spilari.nyr.ruv.is/gql/"

    def _get_live(self):
        url = self.session.http.get(
            self._URL_API_CHANNEL.format(channel=self.match["channel"]),
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "url": validate.url(),
                },
                validate.get("url"),
            ),
        )
        if self.session.http.head(url, raise_for_status=False).status_code >= 400:
            log.error("The content is not available in your region")
            return

        return HLSStream.parse_variant_playlist(self.session, url)

    def _get_vod(self):
        query = f"""
            query getProgram($id: Int!) {{
              Program(id: $id) {{
                title
                episodes(limit: 1, id: {{value: "{self.match["episode"]}"}}) {{
                  title
                  file
                }}
              }}
            }}
        """

        self.author, self.title, url = self.session.http.post(
            self._URL_API_GQL,
            json={
                "operationName": "getProgram",
                "query": dedent(query).strip(),
                "variables": {
                    "id": int(self.match["id"]),
                },
            },
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "data": {
                        "Program": {
                            "episodes": validate.list(
                                {
                                    "file": validate.url(),
                                    "title": str,
                                },
                            ),
                            "title": str,
                        },
                    },
                },
                validate.get(("data", "Program")),
                validate.union_get(
                    "title",
                    ("episodes", 0, "title"),
                    ("episodes", 0, "file"),
                ),
            ),
        )

        if not url.endswith(".m3u8"):
            return {"vod": HTTPStream(self.session, url)}
        else:
            return HLSStream.parse_variant_playlist(self.session, url)

    def _get_streams(self):
        if self.matches["live"]:
            return self._get_live()
        else:
            return self._get_vod()


__plugin__ = Ruv



"""
$description Spanish live TV channels from Atresmedia Television, including Antena 3 and laSexta.
$url atresplayer.com
$type live
$region Spain
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.dash import DASHStream
from streamlink.stream.hls import HLSStream
from streamlink.utils.url import update_scheme


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?atresplayer\.com/directos/.+"),
)
class AtresPlayer(Plugin):
    _channels_api_url = "https://api.atresplayer.com/client/v1/info/channels"
    _player_api_url = "https://api.atresplayer.com/player/v1/live/{channel_id}?NODRM=true"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = update_scheme("https://", f"{self.url.rstrip('/')}/")

    def _get_streams(self):
        channel_path = f"/{self.url.split('/')[-2]}/"
        channel_data = self.session.http.get(
            self._channels_api_url,
            schema=validate.Schema(
                validate.parse_json(),
                [
                    {
                        "id": str,
                        "link": {"url": str},
                    },
                ],
                validate.filter(lambda item: item["link"]["url"] == channel_path),
            ),
        )
        if not channel_data:
            return
        channel_id = channel_data[0]["id"]

        player_api_url = self._player_api_url.format(channel_id=channel_id)
        log.debug(f"Player API URL: {player_api_url}")

        sources = self.session.http.get(
            player_api_url,
            acceptable_status=(200, 403),
            schema=validate.Schema(
                validate.parse_json(),
                validate.any(
                    {
                        "error": str,
                        "error_description": str,
                    },
                    {
                        "sourcesLive": [
                            validate.all(
                                {
                                    "src": validate.url(),
                                    validate.optional("type"): str,
                                },
                                validate.union_get("type", "src"),
                            ),
                        ],
                    },
                ),
            ),
        )
        if "error" in sources:
            log.error(f"Player API error: {sources['error']} - {sources['error_description']}")
            return

        for streamtype, streamsrc in sources.get("sourcesLive"):
            log.debug(f"Stream source: {streamsrc} ({streamtype or 'n/a'})")

            if streamtype == "application/vnd.apple.mpegurl":
                streams = HLSStream.parse_variant_playlist(self.session, streamsrc)
                if not streams:
                    yield "live", HLSStream(self.session, streamsrc)
                else:
                    yield from streams.items()
            elif streamtype == "application/dash+xml":
                yield from DASHStream.parse_manifest(self.session, streamsrc).items()


__plugin__ = AtresPlayer



"""
$description French television news network, covering world news from the European perspective.
$url euronews.com
$type live, vod
$notes Euronews ignores the HTTP client's data decompression capabilities.
$notes Streamlink's optional brotli dependency is therefore strictly required.
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream


log = getLogger(__name__)


@pluginmatcher(
    name="live",
    pattern=re.compile(r"https?://(?:(?P<subdomain>\w+)\.)?euronews\.com/live[?#]?"),
)
@pluginmatcher(
    name="vod",
    pattern=re.compile(r"https?://(?:\w+\.)?euronews\.com/(?!live)\S+"),
)
class Euronews(Plugin):
    API_URL = "https://{subdomain}.euronews.com/api/live/data"

    def _get_live(self):
        log.debug("Querying live API")
        subdomain = self.match["subdomain"] or "www"
        streamtype, streamdata = self.session.http.get(
            self.API_URL.format(subdomain=subdomain),
            params={"locale": subdomain},
            schema=validate.Schema(
                validate.parse_json(),
                validate.any(
                    validate.all(
                        {
                            "videoPrimaryUrl": validate.url(path=validate.endswith(".m3u8")),
                        },
                        validate.transform(lambda data: ("hls", data["videoPrimaryUrl"])),
                    ),
                    validate.all(
                        {
                            "player": "pfp",
                            "videoId": str,
                        },
                        validate.transform(lambda data: ("youtube", data["videoId"])),
                    ),
                ),
            ),
        )

        if streamtype == "hls":
            return HLSStream.parse_variant_playlist(self.session, streamdata)
        if streamtype == "youtube":
            return self.session.streams(f"https://www.youtube.com/watch?v={streamdata}")

    def _get_vod(self):
        log.debug("Looking for VOD")
        videotype, videodata = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.any(
                    validate.all(
                        validate.xml_xpath_string(".//meta[@property='og:video'][1]/@content"),
                        validate.url(),
                        validate.transform(lambda d: ("httpstream", d)),
                    ),
                    validate.all(
                        validate.xml_xpath_string(
                            ".//iframe[@id='pfpPlayer'][starts-with(@src,'https://www.youtube.com/')][1]/@src",
                        ),
                        validate.url(),
                        validate.transform(lambda d: ("plugin", d)),
                    ),
                    validate.all(
                        validate.xml_xpath_string("""
                            .//div
                            [@data-video]
                            [contains(@data-video, //head/meta[@rel='canonical'][@href][1]/@href)]
                            [1]
                            /@data-video
                        """),
                        str,
                        validate.parse_json(),
                        validate.any(
                            validate.all(
                                {"url": validate.url()},
                                validate.transform(lambda d: ("httpstream", d["url"])),
                            ),
                            validate.all(
                                {
                                    "player": "dailymotion",
                                    "id": str,
                                },
                                validate.transform(lambda d: ("plugin", f"https://www.dailymotion.com/embed/video/{d['id']}")),
                            ),
                            validate.all(
                                {
                                    "player": "pfp",
                                    validate.optional("videoId"): str,
                                    validate.optional("id"): str,
                                },
                                validate.transform(lambda d: d.get("videoId") or d.get("id")),
                                str,
                                validate.transform(lambda d: ("plugin", f"https://www.youtube.com/watch?v={d}")),
                            ),
                        ),
                    ),
                    validate.transform(lambda *_: ("error", None)),
                ),
            ),
        )

        if videotype == "httpstream":
            return dict(vod=HTTPStream(self.session, videodata))
        if videotype == "plugin":
            return self.session.streams(videodata)

    def _get_streams(self):
        return self._get_live() if self.matches["live"] else self._get_vod()


__plugin__ = Euronews



"""
$description Video content from Telefe, an Argentine TV station.
$url mitelefe.com
$type live
$metadata title
$region Argentina
"""

import re
from urllib.parse import urlparse

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(re.compile(r"https://mitelefe\.com/(?:telefe-en-)?vivo"))
class Telefe(Plugin):
    _URL_TOKENIZE = "https://mitelefe.com/vidya/tokenize"

    def _get_streams(self):
        self.title, stream_url = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.union((
                    validate.xml_xpath_string(".//meta[@property='og:title'][@content][1]/@content"),
                    validate.all(
                        validate.xml_xpath_string(".//*[@data-player-url][1]/@data-player-url"),
                        validate.none_or_all(
                            validate.url(),
                        ),
                    ),
                )),
            ),
        )
        if not stream_url:
            log.error("Could not find stream URL")
            return None

        parsed = urlparse(stream_url)
        netloc = parsed.netloc.removeprefix("www.")
        if netloc in ("youtube.com", "youtu.be", "dailymotion.com"):
            return self.session.streams(stream_url)

        headers = {
            "Origin": "https://mitelefe.com",
            "Referer": self.url,
        }

        hls_url = self.session.http.post(
            self._URL_TOKENIZE,
            headers={
                "Content-Type": "application/json",
                **headers,
            },
            json={
                "url": stream_url,
            },
            schema=validate.Schema(
                validate.parse_json(),
                validate.any(
                    validate.all(
                        {
                            "url": validate.url(path=validate.endswith(".m3u8")),
                        },
                        validate.get("url"),
                    ),
                    validate.all(
                        str,
                        validate.url(path=validate.endswith(".m3u8")),
                    ),
                ),
            ),
        )
        if self.session.http.head(hls_url, raise_for_status=False).status_code >= 400:
            log.error("Access restricted")
            return None

        return HLSStream.parse_variant_playlist(self.session, hls_url, headers=headers)


__plugin__ = Telefe



"""
$description Russian live-streaming platform for live video game broadcasts.
$url goodgame.ru
$type live
$metadata id
$metadata author
$metadata category
$metadata title
"""

import re
import sys

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    name="default",
    pattern=re.compile(
        r"https?://(?:www\.)?goodgame\.ru/(?P<channel>(?!player)[^/?#]+)",
    ),
)
@pluginmatcher(
    name="player",
    pattern=re.compile(
        r"https?://(?:www\.)?goodgame\.ru/player\?(?P<channel_id>[^&#]+)",
    ),
)
class GoodGame(Plugin):
    _API_PLAYER = "https://goodgame.ru/api/player"
    _API_STREAM = "https://goodgame.ru/api/4/users/{channel}/stream"

    @classmethod
    def stream_weight(cls, stream: str) -> tuple[float, str]:
        if stream == "source":
            return sys.maxsize, stream
        return super().stream_weight(stream)

    def _get_channel_from_player(self):
        return self.session.http.get(
            self._API_PLAYER,
            params={"src": self.match["channel_id"]},
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "streamer_name": str,
                },
                validate.get("streamer_name"),
            ),
        )

    def _api_stream(self, channel):
        return self.session.http.get(
            self._API_STREAM.format(channel=channel),
            acceptable_status=(200, 404),
            schema=validate.Schema(
                validate.parse_json(),
                validate.any(
                    validate.all(
                        {
                            "error": str,
                        },
                        validate.get("error"),
                        validate.transform(lambda data: ("error", data)),
                    ),
                    validate.all(
                        {
                            "online": bool,
                            "id": int,
                            "streamer": {
                                "username": str,
                            },
                            "sources": {
                                str: validate.url(),
                            },
                            "gameObj": {
                                "title": validate.none_or_all(str),
                            },
                            "title": validate.none_or_all(str),
                        },
                        validate.union_get(
                            "online",
                            "id",
                            ("streamer", "username"),
                            ("gameObj", "title"),
                            "title",
                            "sources",
                        ),
                        validate.transform(lambda data: ("data", *data)),
                    ),
                ),
            ),
        )

    def _get_streams(self):
        if self.matches["player"]:
            channel = self._get_channel_from_player()
        else:
            channel = self.match["channel"]

        result, *data = self._api_stream(channel)
        if result == "error":
            log.error(data[0] or "Unknown error")
            return

        online, self.id, self.author, self.category, self.title, sources = data

        if not online:
            return

        streams = {}
        for name, hls_url in sources.items():
            if str.isnumeric(name):
                name = f"{name}p"
            elif name != "source":
                continue
            streams[name] = HLSStream(self.session, hls_url)

        return streams


__plugin__ = GoodGame



from __future__ import annotations

from collections import defaultdict
from itertools import count
from threading import RLock, Thread
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from collections.abc import Iterator


_threadname_lock = RLock()
_threadname_counters: defaultdict[str, Iterator[int]] = defaultdict(count)


class NamedThread(Thread):
    def __init__(self, *args, name: str | None = None, **kwargs):
        with _threadname_lock:
            newname = self.__class__.__name__
            if name:
                newname += f"-{name}"

            # noinspection PyUnresolvedReferences
            kwargs["name"] = f"{newname}-{next(_threadname_counters[newname])}"

        super().__init__(*args, **kwargs)



"""
$description Live TV and video on-demand service owned by Paramount Streaming.
$url pluto.tv
$type live, vod
$metadata id
$metadata author
$metadata category
$metadata title
"""

import re
from dataclasses import dataclass
from typing import ClassVar
from urllib.parse import parse_qsl, urljoin, urlparse
from uuid import uuid4

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, PluginError, pluginmatcher
from streamlink.plugin.api import useragents, validate
from streamlink.stream.hls import HLSSegment, HLSStream, HLSStreamReader, HLSStreamWriter, M3U8Parser
from streamlink.utils.url import update_qsd


log = getLogger(__name__)


@dataclass
class PlutoHLSSegment(HLSSegment):
    ad: bool = False

    _RE_AD: ClassVar[re.Pattern[str]] = re.compile(
        r"""
            _ad(?:/|%2F|_bumper)
            |
            plutotv_filler
        """,
        re.VERBOSE | re.IGNORECASE,
    )

    def __post_init__(self):
        self.ad = self._is_ad()

    def _is_ad(self) -> bool:
        parsed = urlparse(self.uri)

        if parsed.hostname and parsed.hostname.endswith("dai.google.com"):
            return True

        return re.search(self._RE_AD, parsed.path or "") is not None


class PlutoM3U8Parser(M3U8Parser):
    __segment__ = PlutoHLSSegment


class PlutoHLSStreamWriter(HLSStreamWriter):
    def should_filter_segment(self, segment: PlutoHLSSegment):  # type: ignore[override, ty:invalid-method-override]
        return segment.ad or super().should_filter_segment(segment)


class PlutoHLSStreamReader(HLSStreamReader):
    __writer__ = PlutoHLSStreamWriter


class PlutoHLSStream(HLSStream):
    __shortname__ = "hls-pluto"
    __reader__ = PlutoHLSStreamReader
    __parser__ = PlutoM3U8Parser


@pluginmatcher(
    name="live",
    pattern=re.compile(
        r"https?://(?:www\.)?pluto\.tv/(?:\w{2,}/)?live-tv/(?P<id>[^/?]+)",
    ),
)
@pluginmatcher(
    name="series",
    pattern=re.compile(
        r"https?://(?:www\.)?pluto\.tv/(?:\w{2,}/)?on-demand/series/(?P<id_s>[^/]+)(?:/season/\d+)?/episode/(?P<id_e>[^/?]+)",
    ),
)
@pluginmatcher(
    name="movies",
    pattern=re.compile(
        r"https?://(?:www\.)?pluto\.tv/(?:\w{2,}/)?on-demand/movies/(?P<id>[^/?]+)",
    ),
)
class Pluto(Plugin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.session.http.headers.update({"User-Agent": useragents.FIREFOX})
        self._app_version = None
        if not (m := re.search(r"Firefox/(\d+(?:\.\d+)*)", useragents.FIREFOX)):
            raise PluginError("Could not find Firefox version")
        self._device_version = m[1]
        self._client_id = str(uuid4())

    @property
    def app_version(self):
        if self._app_version:
            return self._app_version

        self._app_version = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//head/meta[@name='appVersion']/@content"),
                validate.any(None, str),
            ),
        )
        if not self._app_version:
            raise PluginError("Could not find pluto app version")

        log.debug(f"{self._app_version=}")

        return self._app_version

    def _get_api_data(self, request):
        log.debug(f"_get_api_data: {request=}")

        schema_paths = validate.any(
            validate.all(
                {
                    "paths": [
                        validate.all(
                            {
                                "type": str,
                                "path": str,
                            },
                            validate.union_get("type", "path"),
                        ),
                    ],
                },
                validate.get("paths"),
            ),
            validate.all(
                {
                    "path": str,
                },
                validate.transform(lambda obj: [("hls", obj["path"])]),
            ),
        )
        schema_live = [
            {
                "name": str,
                "id": str,
                "slug": str,
                "stitched": schema_paths,
            },
        ]
        schema_vod = [
            {
                "name": str,
                "id": str,
                "slug": str,
                "genre": str,
                "stitched": validate.any(schema_paths, {}),
                validate.optional("seasons"): [
                    {
                        "episodes": [
                            {
                                "name": str,
                                "_id": str,
                                "slug": str,
                                "stitched": schema_paths,
                            },
                        ],
                    },
                ],
            },
        ]

        return self.session.http.get(
            "https://boot.pluto.tv/v4/start",
            params={
                "appName": "web",
                "appVersion": self.app_version,
                "deviceVersion": self._device_version,
                "deviceModel": "web",
                "deviceMake": "firefox",
                "deviceType": "web",
                "clientID": self._client_id,
                "clientModelNumber": "1.0.0",
                "serverSideAds": "false",
                **request,
            },
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "servers": {
                        "stitcher": validate.url(),
                    },
                    "stitcherParams": str,
                    "sessionToken": str,
                    validate.optional("EPG"): schema_live,
                    validate.optional("VOD"): schema_vod,
                },
            ),
        )

    def _get_streams_live(self):
        data = self._get_api_data({"channelSlug": self.match["id"]})
        epg = data.get("EPG", [])
        media = next((e for e in epg if e["id"] == self.match["id"]), None)
        if not media:
            return

        self.id = media["id"]
        self.title = media["name"]

        return data, media["stitched"]

    def _get_streams_series(self):
        data = self._get_api_data({"seriesIDs": self.match["id_s"]})
        vod = data.get("VOD", [])
        media = next((v for v in vod if v["id"] == self.match["id_s"]), None)
        if not media:
            return
        seasons = media.get("seasons", [])
        episode = next((e for s in seasons for e in s["episodes"] if e["_id"] == self.match["id_e"]), None)
        if not episode:
            return

        self.id = episode["_id"]
        self.author = media["name"]
        self.category = media["genre"]
        self.title = episode["name"]

        return data, episode["stitched"]

    def _get_streams_movies(self):
        data = self._get_api_data({"seriesIDs": self.match["id"]})
        vod = data.get("VOD", [])
        media = next((v for v in vod if v["id"] == self.match["id"]), None)
        if not media:
            return

        self.id = media["id"]
        self.category = media["genre"]
        self.title = media["name"]

        return data, media["stitched"]

    def _get_streams(self):
        res = None
        if self.matches["live"]:
            res = self._get_streams_live()
        elif self.matches["series"]:
            res = self._get_streams_series()
        elif self.matches["movies"]:
            res = self._get_streams_movies()

        if not res:
            return

        data, paths = res
        for mediatype, path in paths:
            if mediatype != "hls":
                continue

            params = dict(parse_qsl(data["stitcherParams"]))
            params["jwt"] = data["sessionToken"]
            params["includeExtendedEvents"] = "true"
            params["masterJWTPassthrough"] = "true"
            url = urljoin(data["servers"]["stitcher"], "v2" + path)
            url = update_qsd(url, params)

            return PlutoHLSStream.parse_variant_playlist(self.session, url)


__plugin__ = Pluto



import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.plugin import LOW_PRIORITY, parse_params
from streamlink.stream.hls import HLSStream
from streamlink.utils.url import update_scheme


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"hls(?:variant)?://(?P<url>\S+)(?:\s(?P<params>.+))?$"),
)
@pluginmatcher(
    priority=LOW_PRIORITY,
    pattern=re.compile(
        # URL with explicit scheme, or URL with implicit HTTPS scheme and a path
        r"(?P<url>[^/]+/\S+\.m3u8(?:\?\S*)?)(?:\s(?P<params>.+))?$",
        re.IGNORECASE,
    ),
)
class HLSPlugin(Plugin):
    def _get_streams(self):
        data = self.match.groupdict()
        url = update_scheme("https://", str(data.get("url", "")), force=False)
        params = parse_params(data.get("params"))
        log.debug(f"URL={url}; params={params}")

        streams = HLSStream.parse_variant_playlist(self.session, url, **params)

        return streams or {"live": HLSStream(self.session, url, **params)}


__plugin__ = HLSPlugin



from __future__ import annotations

from os import environ
from pathlib import Path
from shutil import which
from typing import TYPE_CHECKING

from streamlink.compat import is_darwin, is_win32


if TYPE_CHECKING:
    from collections.abc import Iterable


def _resolve_executable(paths: Iterable[Path], *exes: str) -> Path | None:
    for exe in exes:
        resolved = which(exe)
        if resolved:
            return Path(resolved).resolve()

    checked = set()
    for path in paths:
        for exe in exes:
            fullpath = str(path / exe)
            if fullpath in checked:
                continue
            checked.add(fullpath)
            resolved = which(fullpath)
            if resolved:
                return Path(resolved).resolve()

    return None


def _find_default_player_win32() -> Path | None:
    envvars = "PROGRAMFILES", "PROGRAMFILES(X86)", "PROGRAMW6432"
    subpath = Path() / "VideoLAN" / "VLC"

    return _resolve_executable(
        (
            Path(p) / subpath
            for p in (environ.get(envvar, None) for envvar in envvars)
            if p
        ),
        "vlc.exe",
    )  # fmt: skip


def _find_default_player_darwin() -> Path | None:
    subpath = Path() / "Applications" / "VLC.app" / "Contents" / "MacOS"

    return _resolve_executable(
        [
            Path("/") / subpath,
            Path.home() / subpath,
        ],
        "VLC",
        "vlc",
    )


def _find_default_player_other() -> Path | None:
    return _resolve_executable(
        [],
        "vlc",
    )


def find_default_player() -> Path | None:
    if is_win32:
        return _find_default_player_win32()
    elif is_darwin:
        return _find_default_player_darwin()
    else:
        return _find_default_player_other()



"""
$description Russian live-streaming and video hosting social platform.
$url vk.com
$url vk.ru
$type live, vod
$metadata id
$metadata author
$metadata title
"""

import re
from hashlib import md5
from urllib.parse import parse_qsl, unquote, urlparse, urlunparse

from streamlink.exceptions import NoStreamsError
from streamlink.logger import getLogger
from streamlink.plugin import Plugin, PluginError, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.dash import DASHStream
from streamlink.stream.hls import HLSStream
from streamlink.utils.url import update_qsd


log = getLogger(__name__)


@pluginmatcher(
    name="video",
    pattern=re.compile(r"https?://(?:\w+\.)?vk\.(?:com|ru)/video(?P<id>-?\d+_\d+)"),
)
@pluginmatcher(
    name="default",
    pattern=re.compile(r"https?://(\w+\.)?vk\.(?:com|ru)/(?!video-?\d+_\d+).+"),
)
class VK(Plugin):
    API_URL = "https://vk.com/al_video.php"
    HASH_COOKIE = "hash429"

    def _get_cookies(self):
        def on_response(res, **__):
            if res.headers.get("x-waf-redirect") == "1":
                if not res.headers.get("X-WAF-Backend-Status"):
                    log.debug("Getting WAF cookie")
                    cookie = res.cookies.get(self.HASH_COOKIE)
                    key = md5(cookie.encode("utf-8")).hexdigest()
                    res.headers["Location"] = update_qsd(res.headers["Location"], qsd={"key": key})
                    return res
                elif res.headers.get("X-WAF-Backend-Status") == "challenge_success":
                    self.session.http.cookies.update(res.cookies)
                    return res

        url = urlunparse(urlparse(self.url)._replace(path="", query="", fragment=""))
        self.session.http.get(url, hooks={"response": on_response})

    def follow_vk_redirect(self):
        if self.matches["video"]:
            return

        try:
            parsed_url = urlparse(self.url)
            true_path = next(unquote(v).split("/")[0] for k, v in parse_qsl(parsed_url.query) if k == "z" and len(v) > 0)
            self.url = f"{parsed_url.scheme}://{parsed_url.netloc}/{true_path}"
            if self.matches["video"]:
                return
        except StopIteration:
            pass

        try:
            self.url = self.session.http.get(
                self.url,
                schema=validate.Schema(
                    validate.parse_html(),
                    validate.xml_xpath_string(".//head/meta[@property='og:url'][@content]/@content"),
                    str,
                ),
            )
        except PluginError:
            pass
        if self.matches["video"]:
            return

        raise NoStreamsError

    def _get_streams(self):
        self._get_cookies()
        self.follow_vk_redirect()

        self.id = self.match["id"]
        if not self.id:
            return

        qsd = dict(parse_qsl(urlparse(self.url).query or ""))
        playlist = qsd.get("list", "")

        log.debug(f"{self.id=}, {playlist=}")

        data = self.session.http.post(
            self.API_URL,
            params={"act": "show"},
            data={
                "act": "show",
                "al": "1",
                "list": playlist,
                "video": self.id,
            },
            headers={
                "Referer": self.url,
                "X-Requested-With": "XMLHttpRequest",
            },
            schema=validate.Schema(
                validate.parse_json(),
                {"payload": list},
                validate.get(("payload", -1)),
                list,
                validate.get(-1),
                validate.any(
                    str,
                    validate.all(
                        {"player": {"params": list}},
                        validate.get(("player", "params", 0)),
                        {
                            validate.optional("hls"): validate.any(None, validate.url()),
                            validate.optional("hls_live"): validate.any(None, validate.url()),
                            validate.optional("hls_ondemand"): validate.any(None, validate.url()),
                            validate.optional("dash_live"): validate.any(None, validate.url()),
                            validate.optional("dash_ondemand"): validate.any(None, validate.url()),
                            validate.optional("md_author"): validate.any(None, str),
                            validate.optional("md_title"): validate.any(None, str),
                        },
                    ),
                ),
            ),
        )
        if type(data) is not dict:
            log.error("Video is inaccessible")
            return

        self.author = data.get("md_author")
        self.title = data.get("md_title")

        if hls := data.get("hls_live") or data.get("hls_ondemand") or data.get("hls"):
            return HLSStream.parse_variant_playlist(self.session, hls)

        if dash := data.get("dash_live") or data.get("dash_ondemand"):
            return DASHStream.parse_manifest(self.session, dash)


__plugin__ = VK



"""
$description US live TV channels. OTT service from USTVnow.
$url ustvnow.com
$type live
$account Required, additional subscription required by some streams
"""

import base64
import json
import re
from urllib.parse import urljoin, urlparse
from uuid import uuid4

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, PluginError, pluginargument, pluginmatcher
from streamlink.stream.hls import HLSStream
from streamlink.utils.crypto import AES, SHA256, pad, unpad


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?ustvnow\.com/channel/live/(?P<chname>\w+)"),
)
@pluginargument(
    "username",
    required=True,
    requires=["password"],
    metavar="USERNAME",
    help="Your USTV Now account username",
)
@pluginargument(
    "password",
    required=True,
    sensitive=True,
    metavar="PASSWORD",
    help="Your USTV Now account password",
)
class USTVNow(Plugin):
    _main_js_re = re.compile(r"""src=['"](main\..*\.js)['"]""")
    _enc_key_re = re.compile(r'(?P<key>AES_(?:Key|IV))\s*:\s*"(?P<value>[^"]+)"')

    TENANT_CODE = "ustvnow"
    _api_url = "https://teleupapi.revlet.net/service/api/v1/"
    _token_url = _api_url + "get/token"
    _signin_url = "https://www.ustvnow.com/signin"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._encryption_config = {}
        self._token = None

    @classmethod
    def encrypt_data(cls, data, key, iv):
        rkey = "".join(reversed(key)).encode("utf8")
        riv = "".join(reversed(iv)).encode("utf8")

        fkey = SHA256.new(rkey).hexdigest()[:32].encode("utf8")

        cipher = AES.new(fkey, AES.MODE_CBC, riv)
        encrypted = cipher.encrypt(pad(data, 16, "pkcs7"))
        return base64.b64encode(encrypted)

    @classmethod
    def decrypt_data(cls, data, key, iv):
        rkey = "".join(reversed(key)).encode("utf8")
        riv = "".join(reversed(iv)).encode("utf8")

        fkey = SHA256.new(rkey).hexdigest()[:32].encode("utf8")

        cipher = AES.new(fkey, AES.MODE_CBC, riv)
        decrypted = cipher.decrypt(base64.b64decode(data))
        if decrypted:
            return unpad(decrypted, 16, "pkcs7")
        else:
            return decrypted

    def _get_encryption_config(self, url):
        # find the path to the main.js
        # load the main.js and extract the config
        if not self._encryption_config:
            res = self.session.http.get(url)
            m = self._main_js_re.search(res.text)
            main_js_path = m and m.group(1)
            if main_js_path:
                res = self.session.http.get(urljoin(url, main_js_path))
                self._encryption_config = dict(self._enc_key_re.findall(res.text))

        return self._encryption_config.get("AES_Key"), self._encryption_config.get("AES_IV")

    @property
    def box_id(self):
        if not self.cache.get("box_id"):
            self.cache.set("box_id", str(uuid4()))
        return self.cache.get("box_id")

    def get_token(self):
        if not self._token:
            log.debug("Getting new session token")
            res = self.session.http.get(
                self._token_url,
                params={
                    "tenant_code": self.TENANT_CODE,
                    "box_id": self.box_id,
                    "product": self.TENANT_CODE,
                    "device_id": 5,
                    "display_lang_code": "ENG",
                    "device_sub_type": "",
                    "timezone": "UTC",
                },
            )

            data = res.json()
            if not data["status"]:
                log.error("Token acquisition failed: {details} ({detail})".format(**data["error"]))
                raise PluginError("could not obtain token")

            self._token = data["response"]["sessionId"]
            log.debug(f"New token: {self._token}")

        return self._token

    def api_request(self, path, data, metadata=None):
        key, iv = self._get_encryption_config(self._signin_url)
        post_data = {
            "data": self.encrypt_data(json.dumps(data).encode("utf8"), key, iv).decode("utf8"),
            "metadata": self.encrypt_data(json.dumps(metadata).encode("utf8"), key, iv).decode("utf8"),
        }
        headers = {
            "box-id": self.box_id,
            "session-id": self.get_token(),
            "tenant-code": self.TENANT_CODE,
            "content-type": "application/json",
        }
        res = self.session.http.post(self._api_url + path, data=json.dumps(post_data), headers=headers).json()
        data = {k: v and json.loads(self.decrypt_data(v, key, iv)) for k, v in res.items()}
        return data

    def login(self, username, password):
        log.debug("Trying to login...")
        resp = self.api_request(
            "send",
            {
                "login_id": username,
                "login_key": password,
                "login_mode": "1",
                "manufacturer": "123",
            },
            {"request": "signin"},
        )

        return resp["data"]["status"]

    def _get_streams(self):
        if not self.login(self.get_option("username"), self.get_option("password")):
            log.error("Failed to login, check username and password")
            return

        path = urlparse(self.url).path.strip("/")
        resp = self.api_request("send", {"path": path}, {"request": "page/stream"})
        if not resp["data"]["status"]:
            log.error("Could not find any streams: {code}: {message}".format(**resp["data"]["error"]))
            return

        for stream in resp["data"]["response"]["streams"]:
            if stream["keys"]["licenseKey"]:
                log.warning("Stream possibly protected by DRM")
            yield from HLSStream.parse_variant_playlist(self.session, stream["url"]).items()


__plugin__ = USTVNow



"""
$description Russian live-streaming platform for gaming and esports, owned by VKontakte. Formerly called vkplay.
$url live.vkvideo.ru
$url live.vkplay.ru
$url vkplay.live
$type live, vod
$metadata id
$metadata author
$metadata category
$metadata title
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, PluginError, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:live\.vk(?:video|play)\.ru|vkplay\.live)/(?P<channel_name>\w+)(?:/?$|(?P<vod>/record/[^?#]+))"),
)
class VKvideolive(Plugin):
    API_URL = "https://api.live.vkvideo.ru/v1"

    _WEIGHTS = {
        "quad_hd": 5,
        "full_hd": 4,
        "high": 3,
        "medium": 2,
        "low": 1,
    }

    @classmethod
    def stream_weight(cls, stream: str) -> tuple[float, str]:
        if stream in cls._WEIGHTS:
            return cls._WEIGHTS[stream], "vkvideolive"

        return super().stream_weight(stream)

    def _query_api(self, channel_name, vod):
        schema_data = validate.all(
            {
                validate.optional("category"): validate.all(
                    {
                        "title": str,
                    },
                    validate.get("title"),
                ),
                "title": str,
                "data": validate.any(
                    [
                        validate.all(
                            {
                                "vid": str,
                                "playerUrls": [
                                    validate.all(
                                        {
                                            "type": str,
                                            "url": validate.any("", validate.url()),
                                        },
                                        validate.union_get("type", "url"),
                                    ),
                                ],
                            },
                            validate.union_get("vid", "playerUrls"),
                        ),
                    ],
                    [],
                ),
            },
            validate.union_get(
                "category",
                "title",
                ("data", 0),
            ),
        )
        schema_response = validate.any(
            validate.all(
                {"error": str, "error_description": str},
                validate.get("error_description"),
            ),
            schema_data,
        )

        if not vod:
            schema = schema_response
        else:
            schema = validate.all(
                {
                    "data": {
                        "record": schema_response,
                    },
                },
                validate.get(("data", "record")),
            )

        data = self.session.http.get(
            f"{self.API_URL}/blog/{channel_name}/public_video_stream{vod}",
            headers={"Referer": self.url},
            acceptable_status=(200, 404),
            schema=validate.Schema(
                validate.parse_json(),
                schema,
            ),
        )
        if isinstance(data, str):
            raise PluginError(f"VKvideo API error: {data}")

        return data

    def _get_streams(self):
        self.author = self.match["channel_name"]
        log.debug(f"Channel name: {self.author}")
        vod = self.match["vod"] or ""
        log.debug(f"VOD: {vod}")

        data = self._query_api(channel_name=self.author, vod=vod)

        self.category, self.title, streamdata = data
        if not streamdata:
            return

        self.id, streams = streamdata

        if not vod:
            for streamtype, streamurl in streams:
                if streamurl and streamtype == "live_hls":
                    yield from HLSStream.parse_variant_playlist(self.session, streamurl).items()
        else:
            for streamtype, streamurl in streams:
                if streamurl and streamtype in self._WEIGHTS:
                    yield streamtype, HTTPStream(self.session, streamurl)


__plugin__ = VKvideolive



from __future__ import annotations

from typing import TYPE_CHECKING

from streamlink.utils.formatter import Formatter as _BaseFormatter
from streamlink_cli.utils.path import replace_chars, replace_path, truncate_path


if TYPE_CHECKING:
    from pathlib import Path


class Formatter(_BaseFormatter):
    title = _BaseFormatter.format

    def path(self, pathlike: str, charmap: str | None = None, max_filename_length: int = 255) -> Path:
        def mapper(s):
            return replace_chars(s, charmap)

        def format_part(part: str, isfile: bool) -> str:
            formatted = self._format(part, mapper, {})
            return truncate_path(formatted, max_filename_length, keep_extension=isfile)

        return replace_path(pathlike, format_part)



"""
$description Turkish live TV channel owned by Detelina Media.
$url tv999.bg
$type live
$region Bulgaria
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.utils.url import update_scheme


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?tv999\.bg/live"),
)
class TV999(Plugin):
    title = "TV999"

    def _get_xpath_string(self, url, xpath):
        return self.session.http.get(
            url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(xpath),
                validate.any(None, validate.url()),
            ),
        )

    def _get_streams(self):
        iframe_url = self._get_xpath_string(self.url, ".//iframe[@src]/@src")
        if not iframe_url:
            return
        hls_url = self._get_xpath_string(iframe_url, ".//source[contains(@src,'m3u8')]/@src")
        if not hls_url:
            return
        return {"live": HLSStream(self.session, update_scheme("http://", hls_url))}


__plugin__ = TV999



"""
$description A state/privately owned Russian live TV channel.
$url 1tv.ru
$type live
$metadata title
$region Russia
"""

import random
import re
from urllib.parse import unquote

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.utils.url import update_qsd


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?1tv\.ru/live"),
)
@pluginmatcher(
    re.compile(r"https?://static\.1tv\.ru/eump/(?:embeds|pages)/1tv_live(?:_orbit-plus-4)?\.html"),
)
class OneTV(Plugin):
    def _get_streams(self):
        if "orbit-plus-4" in self.url:
            channel = "1tv-orbit-plus-4"
            self.title = "Первый канал HD (+4)"
        else:
            channel = "1tvch"
            self.title = "Первый канал HD"

        url = self.session.http.get(
            f"https://stream.1tv.ru/api/playlist/{channel}_as_array.json",
            data={"r": random.randint(1, 100000)},
            schema=validate.Schema(
                validate.parse_json(),
                {"hls": [validate.url()]},
                validate.get(("hls", 0)),
            ),
        )

        if not url:
            return

        log.debug(f"{url}")
        if "georestrictions" in url:
            log.error("Stream is geo-restricted")
            return

        hls_session = self.session.http.get(
            "https://stream.1tv.ru/get_hls_session",
            schema=validate.Schema(
                validate.parse_json(),
                {"s": validate.transform(unquote)},
            ),
        )
        url = update_qsd(url, qsd=hls_session, safe="/:")

        return HLSStream.parse_variant_playlist(self.session, url, name_fmt="{pixels}_{bitrate}")


__plugin__ = OneTV



"""
$description Live TV channels and video on-demand service from france.tv, a French public, state-owned broadcaster.
$url france.tv
$url francetvinfo.fr
$type live, vod
$metadata title
$region France, Andorra, Monaco
"""

import re
from urllib.parse import urlparse

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import useragents, validate
from streamlink.stream.dash import DASHStream
from streamlink.stream.hls import HLSStream
from streamlink.utils.times import localnow


log = getLogger(__name__)


@pluginmatcher(
    name="francetv",
    pattern=re.compile(r"https?://(?:[\w-]+\.)?france\.tv/"),
)
@pluginmatcher(
    name="francetvinfofr",
    pattern=re.compile(r"https?://(?:[\w-]+\.)?francetvinfo\.fr/"),
)
class Pluzz(Plugin):
    PLAYER_VERSION = "5.51.35"

    _URL_GEO = "https://geoftv-a.akamaihd.net/ws/edgescape.json"
    _URL_API = "https://k7.ftven.fr/videos/{video_id}"

    _SCHEMA_VIDEOID_FRANCETV = validate.Schema(
        validate.regex(
            re.compile(r"""\\"options\\":\{\\"id\\":\\"(?P<video_id>[\dA-Fa-f-]{36})\\","""),
            method="search",
        ),
        validate.get("video_id"),
    )
    _SCHEMA_VIDEOID_FRANCETVINFOFR = validate.Schema(
        validate.parse_html(),
        validate.any(
            # overseas live stream
            validate.all(
                validate.xml_xpath_string(""".//script[contains(text(),'"data":{content:{player:{id:"')][1]/text()"""),
                str,
                validate.regex(
                    re.compile(r""""data":\{content:\{player:\{id:"(?P<video_id>[\dA-Fa-f-]{36})","""),
                    method="search",
                ),
                validate.get("video_id"),
            ),
            # news article
            validate.all(
                validate.xml_xpath_string(".//*[@id][contains(@class,'francetv-player-wrapper')][1]/@id"),
                str,
            ),
            # videos
            validate.all(
                validate.xml_xpath_string(".//*[@data-id][contains(@class,'magneto')][1]/@data-id"),
                str,
            ),
            validate.transform(lambda *_: None),
        ),
    )

    def _get_video_id(self):
        return self.session.http.get(
            self.url,
            schema=self._SCHEMA_VIDEOID_FRANCETV if self.matches["francetv"] else self._SCHEMA_VIDEOID_FRANCETVINFOFR,
        )

    def _get_streams(self):
        self.session.http.headers.update({
            "User-Agent": useragents.CHROME,
        })
        if not (m := re.compile(r"Chrome/(\d+)").search(useragents.CHROME)):
            return
        CHROME_VERSION = m.group(1)

        if not (video_id := self._get_video_id()):
            return
        log.debug(f"Video ID: {video_id}")

        # Retrieve geolocation data
        country_code = self.session.http.get(
            self._URL_GEO,
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "reponse": {
                        "geo_info": {
                            "country_code": str,
                        },
                    },
                },
                validate.get(("reponse", "geo_info", "country_code")),
            ),
        )
        log.debug(f"Country: {country_code}")

        video_format, token_url, url, self.title = self.session.http.get(
            self._URL_API.format(video_id=video_id),
            params={
                "country_code": country_code,
                "w": 1920,
                "h": 1080,
                "player_version": self.PLAYER_VERSION,
                "domain": urlparse(self.url).netloc,
                "device_type": "mobile",
                "browser": "chrome",
                "browser_version": CHROME_VERSION,
                "os": "ios",
                "gmt": localnow().strftime("%z"),
            },
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "video": {
                        "format": validate.any("dash", "hls"),
                        "token": {
                            "akamai": validate.url(),
                        },
                        "url": validate.url(),
                    },
                    "meta": {
                        "title": str,
                    },
                },
                validate.union_get(
                    ("video", "format"),
                    ("video", "token", "akamai"),
                    ("video", "url"),
                    ("meta", "title"),
                ),
            ),
        )

        video_url = self.session.http.get(
            token_url,
            params={
                "url": url,
            },
            schema=validate.Schema(
                validate.parse_json(),
                {"url": validate.url()},
                validate.get("url"),
            ),
        )

        if video_format == "dash":
            yield from DASHStream.parse_manifest(self.session, video_url).items()
        elif video_format == "hls":
            yield from HLSStream.parse_variant_playlist(self.session, video_url).items()


__plugin__ = Pluzz



"""
$description 24-hour live-streaming world news channel, based in the United States of America.
$url cbsnews.com
$type live
$metadata id
$metadata title
"""

import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?cbsnews\.com/(?:\w+/)?live/?"),
)
class CBSNews(Plugin):
    def _get_streams(self):
        data = self.session.http.get(
            self.url,
            schema=validate.Schema(
                re.compile(r"CBSNEWS\.defaultPayload\s*=\s*(\{.*?})\s*\n"),
                validate.none_or_all(
                    validate.get(1),
                    validate.parse_json(),
                    {
                        "items": [
                            {
                                "id": str,
                                "canonicalTitle": str,
                                "video": validate.url(),
                                "format": "application/x-mpegURL",
                            },
                        ],
                    },
                    validate.get(("items", 0)),
                    validate.union_get("id", "canonicalTitle", "video"),
                ),
            ),
        )
        if not data:
            return

        self.id, self.title, hls_url = data

        return HLSStream.parse_variant_playlist(self.session, hls_url)


__plugin__ = CBSNews



from __future__ import annotations

import sys
from contextlib import contextmanager, suppress
from getpass import getpass
from json import dumps
from typing import TYPE_CHECKING, Any, TextIO

from streamlink_cli.console.stream import ConsoleStatusMessage
from streamlink_cli.utils import JSONEncoder


if TYPE_CHECKING:
    from streamlink_cli.console.stream import ConsoleOutputStream


class ConsoleOutput:
    def __init__(
        self,
        *,
        console_output: ConsoleOutputStream | None = None,
        file_output: TextIO | None = None,
        json: bool = False,
    ):
        self.json: bool = json
        self._console_output: ConsoleOutputStream | None = console_output
        self._supports_status_messages: bool = console_output is not None and console_output.supports_status_messages()
        self._file_output: TextIO | None = file_output

    @property
    def console_output(self) -> ConsoleOutputStream | None:
        return self._console_output

    @console_output.setter
    def console_output(self, console_output: ConsoleOutputStream | None) -> None:
        self._console_output = console_output
        self._supports_status_messages = console_output is not None and console_output.supports_status_messages()

    @property
    def file_output(self) -> TextIO | None:
        return self._file_output

    @file_output.setter
    def file_output(self, file_output: TextIO | None) -> None:
        if file_output is None or file_output.isatty():
            self._file_output = None
        else:
            self._file_output = file_output

    def supports_status_messages(self) -> bool:
        return self._supports_status_messages

    def close(self):
        if self._console_output:  # pragma: no branch
            with suppress(OSError):
                self._console_output.close()
            self._console_output.restore()
        if self._file_output:  # pragma: no branch
            with suppress(OSError):
                self._file_output.close()

    @staticmethod
    def _write(stream: TextIO, msg: str):
        with suppress(OSError):
            stream.write(msg)
            stream.flush()

    def _write_console(self, msg: str):
        if self._console_output is None:
            return
        self._write(self._console_output, msg)

    def _write_file(self, msg: str):
        if self._file_output is None:
            return
        self._write(self._file_output, msg)

    @contextmanager
    def _prompt(self):
        if not sys.stdin or not sys.stdin.isatty():
            raise OSError("No input TTY available")
        if not self._console_output or not self._console_output.isatty():
            raise OSError("No output TTY available")

        try:
            yield
        except OSError:
            raise
        except Exception as err:
            raise OSError(err) from err

    def ask(self, prompt: str) -> str:
        with self._prompt():
            self._write_console(prompt)
            return input().strip()

    def ask_password(self, prompt: str) -> str:
        with self._prompt():
            return getpass(prompt=prompt, stream=self._console_output)

    def msg(self, msg: str) -> None:
        if self.json:
            return
        msg = f"{msg}\n"
        self._write_console(msg)
        self._write_file(msg)

    def msg_status(self, msg: str) -> None:
        if self.json or not self._supports_status_messages:
            return
        self._write_console(ConsoleStatusMessage(msg))

    def msg_json(self, *objs: Any, **keywords: Any) -> None:
        if not self.json:
            return

        out: list | dict
        if objs and isinstance(objs[0], list):
            out = []
            for obj in objs:
                if isinstance(obj, list):
                    out.extend(obj)
                else:
                    if hasattr(obj, "__json__") and callable(obj.__json__):
                        obj = obj.__json__()
                    out.append(obj)
            if keywords:
                out.append(keywords)
        else:
            out = {}
            for obj in objs:
                if hasattr(obj, "__json__") and callable(obj.__json__):
                    obj = obj.__json__()
                if not isinstance(obj, dict):
                    continue
                out.update(**obj)
            out.update(**keywords)

        if self._console_output is not None:
            # don't escape Unicode characters outside the ASCII range if the output encoding is UTF-8
            ensure_ascii = self._console_output.encoding != "utf-8"
            msg = dumps(out, cls=JSONEncoder, ensure_ascii=ensure_ascii, indent=2)
            self._write_console(f"{msg}\n")

        if self._file_output is not None:
            msg = dumps(out, cls=JSONEncoder, ensure_ascii=False, indent=2)
            self._write(self._file_output, f"{msg}\n")



"""
$description Global live streaming and video hosting platform.
$url player.tvibo.com
$type live
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://player\.tvibo\.com/\w+/(?P<id>\d+)"),
)
class Tvibo(Plugin):
    _api_url = "http://panel.tvibo.com/api/player/streamurl/{id}"

    def _get_streams(self):
        channel_id = self.match.group("id")

        api_response = self.session.http.get(self._api_url.format(id=channel_id), acceptable_status=(200, 404))

        data = self.session.http.json(api_response)
        log.trace("%r", data)
        if data.get("st"):
            yield "source", HLSStream(self.session, data["st"])
        elif data.get("error"):
            log.error(data["error"]["message"])


__plugin__ = Tvibo



from streamlink.stream.hls.hls import HLSStream, HLSStreamReader, HLSStreamWorker, HLSStreamWriter, MuxedHLSStream
from streamlink.stream.hls.m3u8 import (
    M3U8,
    M3U8Parser,
    parse_m3u8,
    parse_tag,
)
from streamlink.stream.hls.segment import (
    ByteRange,
    DateRange,
    ExtInf,
    HLSPlaylist,
    HLSSegment,
    IFrameStreamInfo,
    Key,
    Map,
    Media,
    Resolution,
    Start,
    StreamInfo,
)



"""
$description Global live-streaming and video hosting platform for the creative community.
$url picarto.tv
$type live, vod
$metadata author
$metadata category
$metadata title
"""

import re
from textwrap import dedent
from urllib.parse import urlparse

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    name="streampopout",
    pattern=re.compile(r"https?://(?:www\.)?picarto\.tv/streampopout/(?P<po_user>[^/]+)/public$"),
)
@pluginmatcher(
    name="videopopout",
    pattern=re.compile(r"https?://(?:www\.)?picarto\.tv/videopopout/(?P<po_vod_id>\d+)$"),
)
@pluginmatcher(
    name="vod",
    pattern=re.compile(r"https?://(?:www\.)?picarto\.tv/[^/]+/videos/(?P<vod_id>\d+)$"),
)
@pluginmatcher(
    name="user",
    pattern=re.compile(r"https?://(?:www\.)?picarto\.tv/(?P<user>[^/?&]+)$"),
)
class Picarto(Plugin):
    API_URL_LIVE = "https://ptvintern.picarto.tv/api/channel/detail/{username}"
    API_URL_VOD = "https://ptvintern.picarto.tv/ptvapi"
    HLS_URL = "https://{netloc}/stream/hls/{file_name}/index.m3u8"

    def get_live(self, username):
        channel, multistreams, loadbalancer = self.session.http.get(
            self.API_URL_LIVE.format(username=username),
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "channel": validate.any(
                        None,
                        {
                            "id": int,
                            "title": str,
                            "private": bool,
                            "categories": [{"name": str}],
                        },
                    ),
                    "getMultiStreams": validate.any(
                        None,
                        {
                            "streams": [
                                {
                                    "id": int,
                                    "channelId": int,
                                    "name": str,
                                    "online": bool,
                                    "stream_name": str,
                                },
                            ],
                        },
                    ),
                    "getLoadBalancerUrl": validate.any(
                        None,
                        {
                            "url": validate.any(None, validate.transform(lambda url: urlparse(url).netloc)),
                        },
                    ),
                },
                validate.union_get("channel", "getMultiStreams", "getLoadBalancerUrl"),
            ),
        )
        if not channel or not multistreams or not loadbalancer:
            return

        if channel["private"]:
            log.error("This is a private stream")
            return

        user_id = channel["id"]
        if not (stream := next((stream for stream in multistreams["streams"] if stream["channelId"] == user_id), None)):
            log.error("No available stream found in 'multistreams' data")
            return

        self.author = username
        self.category = next(iter(channel["categories"]), {}).get("name")
        self.title = channel["title"]

        hls_url = self.HLS_URL.format(
            netloc=loadbalancer["url"],
            file_name=stream["stream_name"],
        )

        return HLSStream.parse_variant_playlist(self.session, hls_url)

    def get_vod(self, vod_id):
        data = {
            "query": dedent("""
                query ($videoId: ID!) {
                  video(id: $videoId) {
                    id
                    title
                    file_name
                    video_recording_image_url
                    channel {
                      name
                    }
                  }
                }
            """).lstrip(),
            "variables": {"videoId": vod_id},
        }
        vod_data = self.session.http.post(
            self.API_URL_VOD,
            json=data,
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "data": {
                        "video": validate.any(
                            None,
                            {
                                "id": int,
                                "title": str,
                                "file_name": str,
                                "video_recording_image_url": str,
                                "channel": {"name": str},
                            },
                        ),
                    },
                },
                validate.get(("data", "video")),
            ),
        )
        if not vod_data:
            return

        self.author = vod_data["channel"]["name"]
        self.category = "VOD"
        self.title = vod_data["title"]

        netloc = urlparse(vod_data["video_recording_image_url"]).netloc
        hls_url = self.HLS_URL.format(
            netloc=netloc,
            file_name=vod_data["file_name"],
        )

        return HLSStream.parse_variant_playlist(self.session, hls_url)

    def _get_streams(self):
        m = self.match.groupdict()

        if m.get("po_vod_id") or m.get("vod_id"):
            log.debug("Type=VOD")
            return self.get_vod(m.get("po_vod_id") or m.get("vod_id"))
        elif m.get("po_user") or m.get("user"):
            log.debug("Type=Live")
            return self.get_live(m.get("po_user") or m.get("user"))


__plugin__ = Picarto
